# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for cleanup task."""
# pylint: disable=protected-access

import datetime
import unittest

import six

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import appengine_test_utils
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import cleanup
from libs.issue_management import issue_tracker_policy

ISSUE_IGNORE_LABEL = 'ClusterFuzz-Ignore'
ISSUE_INVALID_FUZZER_LABEL = 'ClusterFuzz-Invalid-Fuzzer'
ISSUE_MISTRIAGED_LABEL = 'ClusterFuzz-Wrong'
ISSUE_NEEDS_FEEDBACK_LABEL = 'Needs-Feedback'
ISSUE_VERIFIED_LABEL = 'ClusterFuzz-Verified'
ISSUE_FUZZ_BLOCKER_LABEL = 'Fuzz-Blocker'


@test_utils.with_cloud_emulators('datastore')
class GetPredatorResultItemTest(unittest.TestCase):
  """Tests for the get_predator_result_item helper function."""

  def test_with_components(self):
    """Ensure that we return the components for test cases which have them."""
    result_with_component = {'result': {'suspected_components': ['A', 'B>C']}}

    testcase = test_utils.create_generic_testcase()
    testcase.set_metadata('predator_result', result_with_component)

    actual = cleanup._get_predator_result_item(testcase, 'suspected_components')
    self.assertListEqual(actual, ['A', 'B>C'])

  def test_no_components(self):
    """Ensure that we handle cases with a result, but no components field."""
    result_no_component = {'result': {}}

    testcase = test_utils.create_generic_testcase()
    testcase.set_metadata('predator_result', result_no_component)

    actual = cleanup._get_predator_result_item(testcase, 'suspected_components')
    self.assertIsNone(actual)

  def test_no_result(self):
    """Ensure that we handle cases without a predator result."""
    testcase = test_utils.create_generic_testcase()
    testcase.delete_metadata('predator_result')

    actual = cleanup._get_predator_result_item(
        testcase, 'suspected_components', default=[])
    self.assertListEqual(actual, [])


@test_utils.with_cloud_emulators('datastore')
class CleanupTest(unittest.TestCase):
  """Tests for various cleanup functions."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
        'handlers.cron.cleanup.get_crash_occurrence_platforms',
    ])
    self.mock.utcnow.return_value = test_utils.CURRENT_TIME
    self.issue = appengine_test_utils.create_generic_issue()
    self.policy = issue_tracker_policy.get('test-project')

  def test_mark_duplicate_testcase_as_closed_with_no_issue_1(self):
    """Ensure that a regular bug older than 7 days does not get closed."""
    testcase = test_utils.create_generic_testcase(
        created_days_ago=data_types.DUPLICATE_TESTCASE_NO_BUG_DEADLINE + 1)
    testcase.status = 'Processed'
    testcase.put()
    cleanup.mark_duplicate_testcase_as_closed_with_no_issue(testcase=testcase)
    self.assertTrue(testcase.open)

  def test_mark_duplicate_testcase_as_closed_with_no_issue_2(self):
    """Ensure that a duplicate bug older than 7 days, with an associated
    issue does not get closed."""
    testcase = test_utils.create_generic_testcase(
        created_days_ago=data_types.DUPLICATE_TESTCASE_NO_BUG_DEADLINE + 1)
    testcase.bug_information = str(self.issue.id)
    testcase.status = 'Duplicate'
    testcase.put()
    cleanup.mark_duplicate_testcase_as_closed_with_no_issue(testcase=testcase)
    self.assertTrue(testcase.open)

  def test_mark_duplicate_testcase_as_closed_with_no_issue_3(self):
    """Ensure that a duplicate bug older than 7 days, with no associated
    issue does get closed."""
    testcase = test_utils.create_generic_testcase(
        created_days_ago=data_types.DUPLICATE_TESTCASE_NO_BUG_DEADLINE + 1)
    testcase.bug_information = ''
    testcase.status = 'Duplicate'
    testcase.put()
    cleanup.mark_duplicate_testcase_as_closed_with_no_issue(testcase=testcase)
    self.assertFalse(testcase.open)

  def test_mark_duplicate_testcase_as_closed_with_no_issue_4(self):
    """Ensure that a duplicate bug 7 days old does not get closed."""
    testcase = test_utils.create_generic_testcase(
        created_days_ago=data_types.DUPLICATE_TESTCASE_NO_BUG_DEADLINE)
    testcase.bug_information = ''
    testcase.status = 'Duplicate'
    testcase.put()
    cleanup.mark_duplicate_testcase_as_closed_with_no_issue(testcase=testcase)
    self.assertTrue(testcase.open)

  def test_delete_unreproducible_testcase_with_no_issue_1(self):
    """Ensure that a reproducible bug with no crash in last 7 days, and with an
    associated issue does not get deleted."""
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = False
    testcase.put()
    cleanup.delete_unreproducible_testcase_with_no_issue(testcase=testcase)
    self.assertTrue(test_utils.entity_exists(testcase))

  def test_delete_unreproducible_testcase_with_no_issue_2(self):
    """Ensure that an unreproducible bug with no crash in last 7 days, with an
    associated issue does not get deleted."""
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.delete_unreproducible_testcase_with_no_issue(testcase=testcase)
    self.assertTrue(test_utils.entity_exists(testcase))

  def test_delete_unreproducible_testcase_with_no_issue_3(self):
    """Ensure that an unreproducible bug with no crash in last 7 days, and with
    no associated issue does get deleted."""
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = ''
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.delete_unreproducible_testcase_with_no_issue(testcase=testcase)
    self.assertFalse(test_utils.entity_exists(testcase))

  def test_delete_unreproducible_testcase_with_no_issue_4(self):
    """Ensure that an unreproducible bug with crash in last 7 days does not get
    deleted."""
    self.mock.get_crash_occurrence_platforms.return_value = ['Linux']
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = True
    testcase.bug_information = ''
    testcase.put()
    cleanup.delete_unreproducible_testcase_with_no_issue(testcase=testcase)
    self.assertTrue(test_utils.entity_exists(testcase))

  def test_delete_unreproducible_testcase_with_no_issue_5(self):
    """Ensure that an unreproducible bug created in last 7 days and with crash
    seen in last 7 days does not get deleted."""
    self.mock.get_crash_occurrence_platforms.return_value = ['Linux']
    testcase = test_utils.create_generic_testcase(
        created_days_ago=data_types.UNREPRODUCIBLE_TESTCASE_NO_BUG_DEADLINE - 1)
    testcase.one_time_crasher_flag = True
    testcase.bug_information = ''
    testcase.put()
    cleanup.delete_unreproducible_testcase_with_no_issue(testcase=testcase)
    self.assertTrue(test_utils.entity_exists(testcase))

  def test_mark_issue_as_closed_if_testcase_is_fixed_1(self):
    """Ensure that we don't close issue if associated testcase is open and
    reproducible."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.open = True
    testcase.one_time_crasher_flag = False
    testcase.put()
    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotEqual(self.issue.status, 'Verified')
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_mark_issue_as_closed_if_testcase_is_fixed_2(self):
    """Ensure that we don't close issue if associated testcase is open and
    unreproducible."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.open = True
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotEqual(self.issue.status, 'Verified')
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_mark_issue_as_closed_if_testcase_is_fixed_3(self):
    """Ensure that we close issue if associated testcase is unreproducible, but
    is explicitly marked as closed."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.open = False
    testcase.fixed = 'Yes'
    testcase.put()
    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertEqual(self.issue.status, 'Verified')
    self.assertIn('ClusterFuzz testcase 1 is verified as fixed.',
                  self.issue._monorail_issue.comment)

  def test_mark_issue_as_closed_if_testcase_is_fixed_4(self):
    """Ensure that we close issue if associated testcase is closed and
    reproducible, and the similar open testcase is unreproducible."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.fixed = '1:2'
    testcase.open = False
    testcase.one_time_crasher_flag = False
    testcase.put()

    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.one_time_crasher_flag = True
    similar_testcase.put()

    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertEqual(self.issue.status, 'Verified')
    self.assertIn(
        'ClusterFuzz testcase 1 is verified as fixed in '
        'https://test-clusterfuzz.appspot.com/revisions'
        '?job=test_content_shell_drt&range=1:2',
        self.issue._monorail_issue.comment)

  def test_mark_issue_as_closed_if_testcase_is_fixed_5(self):
    """Ensure that we don't close issue if associated testcase is closed and
    reproducible, but there is a similar testcase is opened."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.fixed = '1:2'
    testcase.open = False
    testcase.one_time_crasher_flag = False
    testcase.put()

    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.one_time_crasher_flag = False
    similar_testcase.put()

    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotEqual(self.issue.status, 'Verified')
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_mark_issue_as_closed_if_testcase_is_fixed_6(self):
    """Ensure that we close issue if all associated testcases are closed and
    reproducible."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.fixed = '1:2'
    testcase.open = False
    testcase.one_time_crasher_flag = False
    testcase.put()

    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.open = False
    similar_testcase.put()

    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertEqual(self.issue.status, 'Verified')
    self.assertIn(
        'ClusterFuzz testcase 1 is verified as fixed in '
        'https://test-clusterfuzz.appspot.com/revisions'
        '?job=test_content_shell_drt&range=1:2',
        self.issue._monorail_issue.comment)

  def test_mark_issue_as_closed_if_testcase_is_fixed_7(self):
    """Ensure that we close issue if issue is marked fixed and all associated
    testcases are closed and reproducible."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.fixed = '1:2'
    testcase.open = False
    testcase.one_time_crasher_flag = False
    testcase.put()

    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.open = False
    similar_testcase.put()

    self.issue._monorail_issue.open = False
    self.issue.status = 'Fixed'

    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertEqual(self.issue.status, 'Verified')
    self.assertIn(
        'ClusterFuzz testcase 1 is verified as fixed in '
        'https://test-clusterfuzz.appspot.com/revisions'
        '?job=test_content_shell_drt&range=1:2',
        self.issue._monorail_issue.comment)

  def test_mark_issue_as_closed_if_testcase_is_fixed_8(self):
    """Ensure that we don't close issue when we already did the issue
    verification once."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.fixed = '1:2'
    testcase.open = False
    testcase.one_time_crasher_flag = False
    testcase.put()

    self.issue.status = 'Assigned'
    self.issue._monorail_issue.comments += [
        appengine_test_utils.create_generic_issue_comment(
            labels=[ISSUE_VERIFIED_LABEL])
    ]
    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotEqual(self.issue.status, 'Verified')
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_mark_issue_as_closed_if_testcase_is_fixed_9(self):
    """Ensure that we don't close issue if a developer has labeled the last
    verification as incorrect."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.fixed = '1:2'
    testcase.open = False
    testcase.one_time_crasher_flag = False
    testcase.put()

    self.issue.status = 'Assigned'
    self.issue._monorail_issue.comments += [
        appengine_test_utils.create_generic_issue_comment(
            labels=[ISSUE_MISTRIAGED_LABEL])
    ]
    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotEqual(self.issue.status, 'Verified')
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_mark_issue_as_closed_if_testcase_is_fixed_10(self):
    """Ensure that we don't close issue when this is unreproducible upload."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.status = 'Unreproducible (trunk)'
    testcase.fixed = ''
    testcase.open = False
    testcase.one_time_crasher_flag = False
    testcase.put()
    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotEqual(self.issue.status, 'Verified')
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_mark_issue_as_closed_if_testcase_is_fixed_11(self):
    """Ensure that we mark issue as verified, but don't close it if job
    definition specifies to skip it."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.fixed = '1:2'
    testcase.open = False
    testcase.one_time_crasher_flag = False
    testcase.put()

    data_types.Job(
        name=testcase.job_type,
        platform='LINUX',
        environment_string=('SKIP_AUTO_CLOSE_ISSUE = True\n')).put()

    cleanup.mark_issue_as_closed_if_testcase_is_fixed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertIn(
        'ClusterFuzz testcase 1 is verified as fixed in '
        'https://test-clusterfuzz.appspot.com/revisions'
        '?job=test_content_shell_drt&range=1:2',
        self.issue._monorail_issue.comment)
    self.assertEqual(self.issue.status, 'Assigned')

  def test_mark_testcase_as_closed_if_issue_is_closed_1(self):
    """Test that we don't do anything if testcase is already closed."""
    testcase = test_utils.create_generic_testcase()
    testcase.open = False
    testcase.put()
    cleanup.mark_testcase_as_closed_if_issue_is_closed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertFalse(testcase.open)

  def test_mark_testcase_as_closed_if_issue_is_closed_2(self):
    """Test that we don't do anything if we are unable to get issue object from
    issue tracker."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    cleanup.mark_testcase_as_closed_if_issue_is_closed(
        policy=self.policy, testcase=testcase, issue=None)
    self.assertTrue(testcase.open)

  def test_mark_testcase_as_closed_if_issue_is_closed_3(self):
    """Test that we don't do anything if there is no associated issue i.e.
    bug_information is not set."""
    testcase = test_utils.create_generic_testcase()
    cleanup.mark_testcase_as_closed_if_issue_is_closed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_testcase_as_closed_if_issue_is_closed_4(self):
    """Test that we don't do anything if issue is still open."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    self.issue._monorail_issue.open = True
    cleanup.mark_testcase_as_closed_if_issue_is_closed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_testcase_as_closed_if_issue_is_closed_5(self):
    """Test that we don't do anything if there is a ignore label on issue."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    self.issue._monorail_issue.open = False
    self.issue._monorail_issue.closed = (
        test_utils.CURRENT_TIME - datetime.timedelta(
            days=data_types.CLOSE_TESTCASE_WITH_CLOSED_BUG_DEADLINE + 1))
    self.issue._monorail_issue.comments += [
        appengine_test_utils.create_generic_issue_comment(
            labels=[ISSUE_IGNORE_LABEL])
    ]
    cleanup.mark_testcase_as_closed_if_issue_is_closed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_testcase_as_closed_if_issue_is_closed_6(self):
    """Test that we don't close testcase if issue is closed <= 2 weeks and
    does not have ignore label."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    self.issue._monorail_issue.open = False
    self.issue._monorail_issue.closed = (
        test_utils.CURRENT_TIME - datetime.timedelta(
            days=data_types.CLOSE_TESTCASE_WITH_CLOSED_BUG_DEADLINE))
    cleanup.mark_testcase_as_closed_if_issue_is_closed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_testcase_as_closed_if_issue_is_closed_7(self):
    """Test that we close testcase if issue is closed longer than 2 weeks and
    does not have ignore label."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    self.issue._monorail_issue.open = False
    self.issue._monorail_issue.closed = (
        test_utils.CURRENT_TIME - datetime.timedelta(
            days=data_types.CLOSE_TESTCASE_WITH_CLOSED_BUG_DEADLINE + 1))
    cleanup.mark_testcase_as_closed_if_issue_is_closed(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertFalse(testcase.open)

  def test_mark_testcase_as_closed_if_job_is_invalid_1(self):
    """Test that we don't close testcase if we have a valid job type."""
    testcase = test_utils.create_generic_testcase()

    jobs = [testcase.job_type]
    cleanup.mark_testcase_as_closed_if_job_is_invalid(
        testcase=testcase, jobs=jobs)
    self.assertTrue(testcase.open)

  def test_mark_testcase_as_closed_if_job_is_invalid_2(self):
    """Test that we close testcase if we don't have a job type."""
    testcase = test_utils.create_generic_testcase()

    jobs = []
    cleanup.mark_testcase_as_closed_if_job_is_invalid(
        testcase=testcase, jobs=jobs)
    self.assertFalse(testcase.open)

  def test_mark_unreproducible_testcase_as_fixed_if_issue_is_closed_1(self):
    """Ensure that a reproducible testcase with no associated issue is not
    marked as Fixed."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = ''
    testcase.one_time_crasher_flag = False
    testcase.put()
    cleanup.mark_unreproducible_testcase_as_fixed_if_issue_is_closed(
        testcase=testcase, issue=None)
    self.assertTrue(testcase.open)

  def test_mark_unreproducible_testcase_as_fixed_if_issue_is_closed_2(self):
    """Ensure that a reproducible testcase with associated issue in open state
    is not marked as Fixed."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = False
    testcase.put()
    cleanup.mark_unreproducible_testcase_as_fixed_if_issue_is_closed(
        testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_unreproducible_testcase_as_fixed_if_issue_is_closed_3(self):
    """Ensure that a reproducible testcase with associated issue in closed state
    is not marked as Fixed."""
    self.issue._monorail_issue.open = False
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = False
    testcase.put()
    cleanup.mark_unreproducible_testcase_as_fixed_if_issue_is_closed(
        testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_unreproducible_testcase_as_fixed_if_issue_is_closed_4(self):
    """Ensure that an unreproducible testcase with associated issue in open
    state is marked as Fixed."""
    self.issue._monorail_issue.open = True
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.mark_unreproducible_testcase_as_fixed_if_issue_is_closed(
        testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_unreproducible_testcase_as_fixed_if_issue_is_closed_5(self):
    """Ensure that an unreproducible testcase with associated issue in closed
    state is marked as Fixed."""
    self.issue._monorail_issue.open = False
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.mark_unreproducible_testcase_as_fixed_if_issue_is_closed(
        testcase=testcase, issue=self.issue)
    self.assertFalse(testcase.open)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_1(
      self):
    """Ensure that a reproducible testcase with no associated issue is not
    closed."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = ''
    testcase.one_time_crasher_flag = False
    testcase.put()
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=None)
    self.assertTrue(testcase.open)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_2(
      self):
    """Ensure that an unreproducible testcase with no associated issue is not
    closed."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = ''
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=None)
    self.assertTrue(testcase.open)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_3(
      self):
    """Ensure that an unreproducible testcase with a closed issue is not
    closed."""
    self.issue._monorail_issue.open = False
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_4(
      self):
    """Ensure that an unreproducible testcase with an open issue and with crash
    still seen in crash stats is not closed."""
    self.issue = appengine_test_utils.create_generic_issue()
    self.mock.get_crash_occurrence_platforms.return_value = ['Linux']
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_5(
      self):
    """Ensure that an unreproducible testcase with an open issue, with crash not
    seen in crash stats, but with other open reproducible testcase is not
    closed."""
    self.issue = appengine_test_utils.create_generic_issue()
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.one_time_crasher_flag = False
    similar_testcase.put()
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_6(
      self):
    """Ensure that an unreproducible testcase with an open issue, with crash not
    seen in crash stats, but with mistriaged issue label is not closed."""
    self.issue = appengine_test_utils.create_generic_issue()
    self.mock.get_crash_occurrence_platforms.return_value = []
    self.issue._monorail_issue.comments += [
        appengine_test_utils.create_generic_issue_comment(
            labels=[ISSUE_MISTRIAGED_LABEL])
    ]
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_7(
      self):
    """Ensure that an unreproducible testcase with an open issue, with crash not
    seen in crash stats, and status as Unreproducible does not lead to closing
    of issue."""
    self.issue = appengine_test_utils.create_generic_issue()
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.status = 'Unreproducible'
    testcase.put()
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)
    self.assertEqual('Assigned', self.issue.status)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_8(
      self):
    """Ensure that an unreproducible testcase with an open issue, created within
    the deadline and crash seen in crash stats is not not closed."""
    self.issue = appengine_test_utils.create_generic_issue()
    self.mock.get_crash_occurrence_platforms.return_value = ['Linux']
    testcase = test_utils.create_generic_testcase(
        created_days_ago=data_types.UNREPRODUCIBLE_TESTCASE_WITH_BUG_DEADLINE -
        1)
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)
    self.assertEqual('Assigned', self.issue.status)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_9(
      self):
    """Ensure that an unreproducible testcase with an open issue, with crash not
    seen in crash stats, is closed."""
    self.issue = appengine_test_utils.create_generic_issue()
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertFalse(testcase.open)
    self.assertEqual('WontFix', self.issue.status)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_10(
      self):
    """Ensure that an unreproducible testcase with an open issue, with crash not
    seen in crash stats, but with an uploader email is not closed."""
    self.issue = appengine_test_utils.create_generic_issue()
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.uploader_email = 'abc@example.com'
    testcase.put()
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)
    self.assertEqual('Assigned', self.issue.status)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_11(
      self):
    """Ensure that an unreproducible testcase with an open issue, with crash not
    seen in crash stats, but reproduced yesterday as as part of progression task
    is not closed."""
    self.issue = appengine_test_utils.create_generic_issue()
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    testcase.set_metadata('last_tested_crash_time', test_utils.CURRENT_TIME)
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertTrue(testcase.open)
    self.assertEqual('Assigned', self.issue.status)

  def test_mark_unreproducible_testcase_and_issue_as_closed_after_deadline_12(
      self):
    """Ensure that an unreproducible testcase with an open issue, with crash not
    seen in crash stats and progression task is closed."""
    self.issue = appengine_test_utils.create_generic_issue()
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.one_time_crasher_flag = True
    testcase.put()
    testcase.set_metadata(
        'last_tested_crash_time',
        test_utils.CURRENT_TIME - datetime.timedelta(
            days=data_types.UNREPRODUCIBLE_TESTCASE_WITH_BUG_DEADLINE + 1))
    cleanup.mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertFalse(testcase.open)
    self.assertEqual('WontFix', self.issue.status)

  def test_notify_closed_issue_if_testcase_is_open_1(self):
    """Test that we don't do anything if testcase is already closed."""
    testcase = test_utils.create_generic_testcase()
    testcase.open = False
    testcase.put()
    cleanup.notify_closed_issue_if_testcase_is_open(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotIn(ISSUE_NEEDS_FEEDBACK_LABEL, self.issue.labels)

  def test_notify_closed_issue_if_testcase_is_open_2(self):
    """Test that we don't do anything if testcase has status unreproducible
    (upload didn't reproduce)."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.status = 'Unreproducible'
    testcase.put()
    self.issue._monorail_issue.open = False
    self.issue._monorail_issue.closed = (
        test_utils.CURRENT_TIME - datetime.timedelta(
            days=data_types.NOTIFY_CLOSED_BUG_WITH_OPEN_TESTCASE_DEADLINE + 1))
    cleanup.notify_closed_issue_if_testcase_is_open(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotIn(ISSUE_NEEDS_FEEDBACK_LABEL, self.issue.labels)

  def test_notify_closed_issue_if_testcase_is_open_3(self):
    """Test that we don't do anything if we are unable to get issue object from
    issue tracker."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    cleanup.notify_closed_issue_if_testcase_is_open(
        policy=self.policy, testcase=testcase, issue=None)
    self.assertNotIn(ISSUE_NEEDS_FEEDBACK_LABEL, self.issue.labels)

  def test_notify_closed_issue_if_testcase_is_open_4(self):
    """Test that we don't do anything if there is no associated issue i.e.
    bug_information is not set."""
    testcase = test_utils.create_generic_testcase()
    cleanup.notify_closed_issue_if_testcase_is_open(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotIn(ISSUE_NEEDS_FEEDBACK_LABEL, self.issue.labels)

  def test_notify_closed_issue_if_testcase_is_open_5(self):
    """Test that we don't do anything if issue is still open."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    self.issue._monorail_issue.open = True
    cleanup.notify_closed_issue_if_testcase_is_open(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotIn(ISSUE_NEEDS_FEEDBACK_LABEL, self.issue.labels)

  def test_notify_closed_issue_if_testcase_is_open_6(self):
    """Test that we don't do anything if we have not exceeded the notification
    deadline."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    self.issue._monorail_issue.open = False
    self.issue._monorail_issue.closed = (
        test_utils.CURRENT_TIME - datetime.timedelta(
            days=data_types.NOTIFY_CLOSED_BUG_WITH_OPEN_TESTCASE_DEADLINE))
    cleanup.notify_closed_issue_if_testcase_is_open(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotIn(ISSUE_NEEDS_FEEDBACK_LABEL, self.issue.labels)

  def test_notify_closed_issue_if_testcase_is_open_7(self):
    """Test that we don't do anything if there is an ignore label already on
    the issue."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    self.issue._monorail_issue.open = False
    self.issue._monorail_issue.closed = (
        test_utils.CURRENT_TIME - datetime.timedelta(
            days=data_types.NOTIFY_CLOSED_BUG_WITH_OPEN_TESTCASE_DEADLINE + 1))
    self.issue._monorail_issue.comments += [
        appengine_test_utils.create_generic_issue_comment(
            labels=[ISSUE_IGNORE_LABEL])
    ]
    cleanup.notify_closed_issue_if_testcase_is_open(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotIn(ISSUE_NEEDS_FEEDBACK_LABEL, self.issue.labels)

  def test_notify_closed_issue_if_testcase_is_open_8(self):
    """Test that we don't do anything if there is an needs feedback label
    already on the issue."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    self.issue._monorail_issue.open = False
    self.issue._monorail_issue.closed = (
        test_utils.CURRENT_TIME - datetime.timedelta(
            days=data_types.NOTIFY_CLOSED_BUG_WITH_OPEN_TESTCASE_DEADLINE + 1))
    self.issue._monorail_issue.comments += [
        appengine_test_utils.create_generic_issue_comment(
            labels=[ISSUE_NEEDS_FEEDBACK_LABEL])
    ]
    cleanup.notify_closed_issue_if_testcase_is_open(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertNotIn(ISSUE_NEEDS_FEEDBACK_LABEL, self.issue.labels)

  def test_notify_closed_issue_if_testcase_is_open_9(self):
    """Test that we add notification if we are past the notification deadline
    and we have not added a needs feedback already."""
    testcase = test_utils.create_generic_testcase()
    testcase.bug_information = str(self.issue.id)
    testcase.put()
    self.issue._monorail_issue.open = False
    self.issue._monorail_issue.closed = (
        test_utils.CURRENT_TIME - datetime.timedelta(
            days=data_types.NOTIFY_CLOSED_BUG_WITH_OPEN_TESTCASE_DEADLINE + 1))
    cleanup.notify_closed_issue_if_testcase_is_open(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertIn(ISSUE_NEEDS_FEEDBACK_LABEL, self.issue.labels)

  def test_mark_na_testcase_issues_as_wontfix(self):
    """Test that issue for fixed == 'NA' testcases are closed."""
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.status = 'Processed'
    testcase.open = False
    testcase.fixed = 'NA'
    testcase.bug_information = str(self.issue.id)
    testcase.put()

    cleanup.mark_na_testcase_issues_as_wontfix(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertIn(
        'ClusterFuzz testcase 1 is closed as invalid, so closing issue.',
        self.issue._monorail_issue.comment)
    self.assertEqual('WontFix', self.issue.status)

  def test_mark_na_testcase_issues_as_wontfix_testcase_open(self):
    """Test that valid open testcases don't get their issues closed."""
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.status = 'Processed'
    testcase.bug_information = str(self.issue.id)
    testcase.put()

    cleanup.mark_na_testcase_issues_as_wontfix(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertEqual('Assigned', self.issue.status)

  def test_mark_na_testcase_issues_as_wontfix_still_occurring(self):
    """Test that issue for fixed == 'NA' testcases are not closed if the crash
    is still occurring."""
    self.mock.get_crash_occurrence_platforms.return_value = ['Linux']
    testcase = test_utils.create_generic_testcase()
    testcase.status = 'Processed'
    testcase.open = False
    testcase.fixed = 'NA'
    testcase.bug_information = str(self.issue.id)
    testcase.put()

    cleanup.mark_na_testcase_issues_as_wontfix(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertEqual('Assigned', self.issue.status)

  def test_mark_na_testcase_issues_as_wontfix_similar_testcase(self):
    """Test that issue for fixed == 'NA' testcases are not closed if there is a
    similar testcase attached to the same issue."""
    self.mock.get_crash_occurrence_platforms.return_value = []
    testcase = test_utils.create_generic_testcase()
    testcase.status = 'Processed'
    testcase.open = False
    testcase.fixed = 'NA'
    testcase.bug_information = str(self.issue.id)
    testcase.put()

    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.one_time_crasher_flag = False
    similar_testcase.put()

    cleanup.mark_na_testcase_issues_as_wontfix(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertEqual('Assigned', self.issue.status)

  def test_mark_na_testcase_issues_as_wontfix_mistriaged(self):
    """Test that issue for fixed == 'NA' testcases are not closed if the issue
    was marked as being mistriaged."""
    self.mock.get_crash_occurrence_platforms.return_value = []
    self.issue._monorail_issue.comments += [
        appengine_test_utils.create_generic_issue_comment(
            labels=[ISSUE_MISTRIAGED_LABEL])
    ]
    testcase = test_utils.create_generic_testcase()
    testcase.status = 'Processed'
    testcase.open = False
    testcase.fixed = 'NA'
    testcase.bug_information = str(self.issue.id)
    testcase.put()

    cleanup.mark_na_testcase_issues_as_wontfix(
        policy=self.policy, testcase=testcase, issue=self.issue)
    self.assertEqual('Assigned', self.issue.status)


@test_utils.with_cloud_emulators('datastore')
class UpdateOsLabelsTest(unittest.TestCase):
  """Test updateOsLabels."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.metrics.crash_stats.get',
    ])
    self.policy = issue_tracker_policy.get('test-project')

  def test_no_issue(self):
    """Test no issue."""
    testcase = data_types.Testcase(
        crash_type='type',
        crash_state='state',
        security_flag=True,
        project_name='project')
    testcase.put()
    cleanup.update_os_labels(self.policy, testcase, None)

  def test_labels_added(self):
    """Test os label added from crash stats."""
    testcase = data_types.Testcase(
        crash_type='type',
        crash_state='state',
        security_flag=True,
        project_name='project')
    testcase.put()

    testcase_variant = data_types.TestcaseVariant(
        testcase_id=testcase.key.id(),
        job_type='mac_job',
        is_similar=True,
        platform='mac')
    testcase_variant.put()

    history = data_types.BuildCrashStatsJobHistory(end_time_in_hours=10000)
    history.put()

    rows = [{
        'groups': [{
            'name': 'windows'
        }, {
            'name': 'linux'
        }, {
            'name': 'android:test'
        }, {
            'name': 'android:test2'
        }]
    }]
    self.mock.get.return_value = (1, rows)

    issue = appengine_test_utils.create_generic_issue()
    issue._monorail_issue.labels = []
    cleanup.update_os_labels(self.policy, testcase, issue)
    self.assertEqual({'OS-Windows', 'OS-Linux', 'OS-Mac', 'OS-Android'},
                     set(issue.labels))
    self.mock.get.assert_called_once_with(
        end=10000,
        block='day',
        days=1,
        group_by='platform',
        where_clause=('crash_type = "type" AND crash_state = "state" AND '
                      'security_flag = true AND project = "project"'),
        group_having_clause='',
        sort_by='total_count',
        offset=0,
        limit=1)

  def test_labels_not_added(self):
    """Test os labels not added from crash stats."""
    testcase = data_types.Testcase(
        crash_type='type',
        crash_state='state',
        security_flag=True,
        project_name='project')
    testcase.put()

    history = data_types.BuildCrashStatsJobHistory(end_time_in_hours=10000)
    history.put()

    rows = [{
        'groups': [{
            'name': 'windows'
        }, {
            'name': 'linux'
        }, {
            'name': 'mac'
        }, {
            'name': 'android:test'
        }, {
            'name': 'android:test2'
        }]
    }]
    self.mock.get.return_value = (1, rows)

    issue = appengine_test_utils.create_generic_issue()
    issue._monorail_issue.labels = []
    comment = appengine_test_utils.create_generic_issue_comment(
        labels=['OS-Mac', 'OS-Android'])
    issue._monorail_issue.comments.append(comment)
    issue.labels.add('OS-Windows')

    cleanup.update_os_labels(self.policy, testcase, issue)
    self.assertEqual({'OS-Windows', 'OS-Linux'}, set(issue.labels))


@test_utils.with_cloud_emulators('datastore')
class GetJobsAndPlatformsForProjectTest(unittest.TestCase):
  """Test get_jobs_and_platforms_for_project."""

  def setUp(self):
    data_types.Job(
        name='job1',
        platform='LINUX',
        environment_string=('EXPERIMENTAL = True\nPROJECT_NAME=project1'),
    ).put()
    data_types.Job(
        name='job2',
        platform='MAC',
        environment_string=('CUSTOM_BINARY = True\nPROJECT_NAME=project2'),
    ).put()
    data_types.Job(
        name='job3',
        platform='WINDOWS',
        environment_string=(
            'SYSTEM_BINARY_DIR = C:\\Program Files\\Internet Explorer\\\n'
            'PROJECT_NAME=project3'),
    ).put()
    data_types.Job(
        name='job4',
        platform='ANDROID',
        environment_string=(
            'EXCLUDE_FROM_TOP_CRASHES = True\nPROJECT_NAME=project4'),
    ).put()
    data_types.Job(
        name='job5',
        platform='LINUX',
        environment_string=('PROJECT_NAME=project5')).put()
    data_types.Job(
        name='job6',
        platform='MAC',
        environment_string=('PROJECT_NAME=project6')).put()

  def test(self):
    actual_projects_map = cleanup.get_jobs_and_platforms_for_project()
    expected_projects_map = {
        'project5': cleanup.ProjectMap(set(['job5']), set(['LINUX'])),
        'project6': cleanup.ProjectMap(set(['job6']), set(['MAC']))
    }
    self.assertEqual(actual_projects_map, expected_projects_map)


@test_utils.with_cloud_emulators('datastore')
class GetTopCrashesForAllProjectsAndPlatforms(unittest.TestCase):
  """Test get_top_crashes_for_all_projects_and_platforms."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.metrics.crash_stats.get',
        'clusterfuzz._internal.metrics.crash_stats.get_last_successful_hour',
    ])

    self.top_crashes_rows = [
        {
            'crashState': 'state1',
            'crashType': 'type1',
            'isSecurity': True,
            'totalCount': 350
        },
        {
            'crashState': 'state2',
            'crashType': 'type2',
            'isSecurity': False,
            'totalCount': 450
        },
        {
            'crashState': 'state3',
            'crashType': 'type3',
            'isSecurity': False,
            'totalCount': 250
        },
    ]
    self.mock.get.return_value = (1, self.top_crashes_rows)
    self.mock.get_last_successful_hour.return_value = 10000
    data_types.Job(
        name='job',
        platform='LINUX',
        environment_string=('PROJECT_NAME = project')).put()

  def test(self):
    """Test."""
    expected_top_crashes_map = {
        u'project': {
            'LINUX': [{
                'crashState': 'state1',
                'crashType': 'type1',
                'isSecurity': True,
                'totalCount': 350
            }, {
                'crashState': 'state2',
                'crashType': 'type2',
                'isSecurity': False,
                'totalCount': 450
            }]
        }
    }
    actual_top_crashes_map = (
        cleanup.get_top_crashes_for_all_projects_and_platforms())
    self.assertEqual(expected_top_crashes_map, actual_top_crashes_map)
    self.mock.get.assert_called_once_with(
        end=10000,
        block='day',
        days=7,
        group_by='platform',
        where_clause=('crash_type NOT IN UNNEST('
                      '["Out-of-memory", "Stack-overflow", "Timeout"]) AND '
                      'crash_state NOT IN UNNEST(["NULL"]) AND '
                      'job_type IN UNNEST(["job"]) AND '
                      'platform LIKE "linux%" AND '
                      'project = "project"'),
        group_having_clause='',
        sort_by='total_count',
        offset=0,
        limit=5)


@test_utils.with_cloud_emulators('datastore')
class UpdateTopCrashLabelsTest(unittest.TestCase):
  """Test update_fuzz_blocker_label."""

  def setUp(self):
    helpers.patch_environ(self)

    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.is_oss_fuzz',
        'clusterfuzz._internal.base.utils.is_chromium',
        'clusterfuzz._internal.chrome.build_info.get_release_milestone',
    ])
    self.mock.get_release_milestone.return_value = 63

    self.issue = appengine_test_utils.create_generic_issue()
    self.testcase = test_utils.create_generic_testcase()
    self.mock.is_chromium.return_value = True
    self.mock.is_oss_fuzz.return_value = False
    self.policy = issue_tracker_policy.get('test-project')

  def test_no_top_crashes(self):
    """Test no label is added if there are no top crashes."""
    top_crashes_by_project_and_platform_map = {u'project': {'LINUX': []}}

    cleanup.update_fuzz_blocker_label(self.policy, self.testcase, self.issue,
                                      top_crashes_by_project_and_platform_map)
    self.assertNotIn(ISSUE_FUZZ_BLOCKER_LABEL, self.issue.labels)
    self.assertNotIn(data_types.CHROMIUM_ISSUE_RELEASEBLOCK_BETA_LABEL,
                     self.issue.labels)
    self.assertNotIn('M-63', self.issue.labels)
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_top_crashes_no_match(self):
    """Test no label is added if there are no matching top crashes."""
    top_crashes_by_project_and_platform_map = {
        u'project': {
            'LINUX': [{
                'crashState': 'state1',
                'crashType': 'type1',
                'isSecurity': True,
                'totalCount': 500
            }]
        }
    }
    cleanup.update_fuzz_blocker_label(self.policy, self.testcase, self.issue,
                                      top_crashes_by_project_and_platform_map)
    self.assertNotIn(ISSUE_FUZZ_BLOCKER_LABEL, self.issue.labels)
    self.assertNotIn(data_types.CHROMIUM_ISSUE_RELEASEBLOCK_BETA_LABEL,
                     self.issue.labels)
    self.assertNotIn('M-63', self.issue.labels)
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_top_crashes_with_testcase_closed(self):
    """Test label is not added if testcase is closed."""
    self.testcase.open = False
    self.testcase.put()

    top_crashes_by_project_and_platform_map = {
        u'project': {
            'LINUX': [{
                'crashState': self.testcase.crash_state,
                'crashType': self.testcase.crash_type,
                'isSecurity': self.testcase.security_flag,
                'totalCount': 350
            }]
        }
    }
    cleanup.update_fuzz_blocker_label(self.policy, self.testcase, self.issue,
                                      top_crashes_by_project_and_platform_map)
    self.assertNotIn(ISSUE_FUZZ_BLOCKER_LABEL, self.issue.labels)
    self.assertNotIn(data_types.CHROMIUM_ISSUE_RELEASEBLOCK_BETA_LABEL,
                     self.issue.labels)
    self.assertNotIn('M-63', self.issue.labels)
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_top_crashes_match_single_platform(self):
    """Test label is added if there is a matching top crash."""
    top_crashes_by_project_and_platform_map = {
        u'project': {
            'LINUX': [{
                'crashState': self.testcase.crash_state,
                'crashType': self.testcase.crash_type,
                'isSecurity': self.testcase.security_flag,
                'totalCount': 350
            }]
        }
    }
    self.issue.labels.add('M-62')
    cleanup.update_fuzz_blocker_label(self.policy, self.testcase, self.issue,
                                      top_crashes_by_project_and_platform_map)
    self.assertIn(ISSUE_FUZZ_BLOCKER_LABEL, self.issue.labels)
    self.assertIn(data_types.CHROMIUM_ISSUE_RELEASEBLOCK_BETA_LABEL,
                  self.issue.labels)
    self.assertIn('M-63', self.issue.labels)
    self.assertNotIn('M-62', self.issue.labels)
    self.assertEqual(
        'This crash occurs very frequently on linux platform and is likely '
        'preventing the fuzzer fuzzer1 from making much progress. '
        'Fixing this will allow more bugs to be found.'
        '\n\nMarking this bug as a blocker for next Beta release.'
        '\n\nIf this is incorrect, please add the ClusterFuzz-Wrong label and '
        'remove the ReleaseBlock-Beta label.',
        self.issue._monorail_issue.comment)

  def test_top_crashes_match_single_platform_oss_fuzz(self):
    """Test label is added if there is a matching top crash for external
    project."""
    self.mock.is_oss_fuzz.return_value = True
    self.testcase.set_metadata('fuzzer_binary_name', 'fuzz_target1')
    top_crashes_by_project_and_platform_map = {
        u'project': {
            'LINUX': [{
                'crashState': self.testcase.crash_state,
                'crashType': self.testcase.crash_type,
                'isSecurity': self.testcase.security_flag,
                'totalCount': 350
            }]
        }
    }
    cleanup.update_fuzz_blocker_label(self.policy, self.testcase, self.issue,
                                      top_crashes_by_project_and_platform_map)
    self.assertIn(ISSUE_FUZZ_BLOCKER_LABEL, self.issue.labels)
    self.assertNotIn(data_types.CHROMIUM_ISSUE_RELEASEBLOCK_BETA_LABEL,
                     self.issue.labels)
    self.assertNotIn('M-63', self.issue.labels)
    self.assertEqual(
        'This crash occurs very frequently on linux platform and is likely '
        'preventing the fuzzer fuzz_target1 from making much progress. '
        'Fixing this will allow more bugs to be found.'
        '\n\nIf this is incorrect, please file a bug on '
        'https://github.com/google/oss-fuzz/issues/new',
        self.issue._monorail_issue.comment)

  def test_top_crashes_match_multiple_platforms(self):
    """Test label is added if there is a matching top crash."""
    top_crashes_by_project_and_platform_map = {
        u'project': {
            'LINUX': [{
                'crashState': self.testcase.crash_state,
                'crashType': self.testcase.crash_type,
                'isSecurity': self.testcase.security_flag,
                'totalCount': 500
            }],
            'MAC': [{
                'crashState': self.testcase.crash_state,
                'crashType': self.testcase.crash_type,
                'isSecurity': self.testcase.security_flag,
                'totalCount': 600
            }],
            'WINDOWS': [{
                'crashState': self.testcase.crash_state,
                'crashType': self.testcase.crash_type,
                'isSecurity': self.testcase.security_flag,
                'totalCount': 700
            }]
        }
    }
    cleanup.update_fuzz_blocker_label(self.policy, self.testcase, self.issue,
                                      top_crashes_by_project_and_platform_map)
    self.assertIn(ISSUE_FUZZ_BLOCKER_LABEL, self.issue.labels)
    self.assertIn(data_types.CHROMIUM_ISSUE_RELEASEBLOCK_BETA_LABEL,
                  self.issue.labels)
    self.assertIn('M-63', self.issue.labels)
    self.assertEqual(
        'This crash occurs very frequently on linux, mac and windows platforms '
        'and is likely preventing the fuzzer fuzzer1 from making much '
        'progress. Fixing this will allow more bugs to be found.'
        '\n\nMarking this bug as a blocker for next Beta release.'
        '\n\nIf this is incorrect, please add the ClusterFuzz-Wrong label and '
        'remove the ReleaseBlock-Beta label.',
        self.issue._monorail_issue.comment)

  def test_top_crashes_match_and_label_removed(self):
    """Test label is not added if it was added before and removed."""
    top_crashes_by_project_and_platform_map = {
        u'project': {
            'LINUX': [{
                'crashState': self.testcase.crash_state,
                'crashType': self.testcase.crash_type,
                'isSecurity': self.testcase.security_flag,
                'totalCount': 500
            }]
        }
    }

    self.issue._monorail_issue.comments += [
        appengine_test_utils.create_generic_issue_comment(
            labels=[ISSUE_FUZZ_BLOCKER_LABEL])
    ]
    cleanup.update_fuzz_blocker_label(self.policy, self.testcase, self.issue,
                                      top_crashes_by_project_and_platform_map)
    self.assertNotIn(ISSUE_FUZZ_BLOCKER_LABEL, self.issue.labels)
    self.assertNotIn(data_types.CHROMIUM_ISSUE_RELEASEBLOCK_BETA_LABEL,
                     self.issue.labels)
    self.assertNotIn('M-63', self.issue.labels)
    self.assertEqual('', self.issue._monorail_issue.comment)


@test_utils.with_cloud_emulators('datastore')
class UpdateComponentsTest(unittest.TestCase):
  """Tests for update_component_labels."""

  def setUp(self):
    self.issue = appengine_test_utils.create_generic_issue()
    self.testcase = test_utils.create_generic_testcase()

  def test_components_added(self):
    """Ensure that we add components when applicable."""
    self.testcase.set_metadata(
        'predator_result', {'result': {
            'suspected_components': ['A', 'B>C']
        }})

    cleanup.update_component_labels(self.testcase, self.issue)
    self.assertIn('A', self.issue.components)
    self.assertIn('B>C', self.issue.components)
    self.assertIn('Test-Predator-Auto-Components', self.issue.labels)

  def test_components_not_reapplied(self):
    """Ensure that we don't re-add components once applied."""
    self.testcase.set_metadata(
        'predator_result', {'result': {
            'suspected_components': ['A', 'B>C']
        }})

    comment = appengine_test_utils.create_generic_issue_comment(
        labels=['Test-Predator-Auto-Components'])
    self.issue._monorail_issue.comments.append(comment)

    cleanup.update_component_labels(self.testcase, self.issue)
    self.assertNotIn('A', self.issue.components)
    self.assertNotIn('B>C', self.issue.components)
    self.assertNotIn('Test-Predator-Auto-Components', self.issue.labels)

  def test_no_label_added_for_no_components(self):
    """Ensure that we don't add label when there is no component in result."""
    self.testcase.set_metadata('predator_result', {})
    self.issue.components.add('A')
    cleanup.update_component_labels(self.testcase, self.issue)
    self.assertIn('A', self.issue.components)
    self.assertNotIn('Test-Predator-Auto-Components', self.issue.labels)

  def test_no_label_added_for_same_components(self):
    """Ensure that we don't add label when there is no component in result."""
    self.testcase.set_metadata(
        'predator_result', {'result': {
            'suspected_components': ['A', 'B>C']
        }})
    self.issue.components.add('A')
    self.issue.components.add('B>C')
    self.issue.components.add('D')
    cleanup.update_component_labels(self.testcase, self.issue)
    self.assertIn('A', self.issue.components)
    self.assertIn('B>C', self.issue.components)
    self.assertIn('D', self.issue.components)
    self.assertNotIn('Test-Predator-Auto-Components', self.issue.labels)

  def test_no_label_added_for_more_specific_component(self):
    """Ensure that we don't add label when there is a more specific component
    already."""
    self.testcase.set_metadata('predator_result',
                               {'result': {
                                   'suspected_components': ['A']
                               }})
    self.issue.components.add('A>B')
    self.issue.components.add('D')
    cleanup.update_component_labels(self.testcase, self.issue)
    self.assertNotIn('A', self.issue.components)
    self.assertIn('A>B', self.issue.components)
    self.assertIn('D', self.issue.components)
    self.assertNotIn('Test-Predator-Auto-Components', self.issue.labels)

  def test_label_added_for_more_specific_component_and_new_component(self):
    """Ensure that we add label when there is a more specific component
    already, but also a new components."""
    self.testcase.set_metadata('predator_result',
                               {'result': {
                                   'suspected_components': ['A', 'E']
                               }})
    self.issue.components.add('A>B')
    self.issue.components.add('D')
    cleanup.update_component_labels(self.testcase, self.issue)
    self.assertNotIn('A', self.issue.components)
    self.assertIn('A>B', self.issue.components)
    self.assertIn('D', self.issue.components)
    self.assertIn('E', self.issue.components)
    self.assertIn('Test-Predator-Auto-Components', self.issue.labels)

  def test_label_added_for_unrelated_component(self):
    """Ensure that we add label when there is a unrelated component with same
    prefix."""
    self.testcase.set_metadata('predator_result',
                               {'result': {
                                   'suspected_components': ['A']
                               }})
    self.issue.components.add('AA>B')
    self.issue.components.add('D')
    cleanup.update_component_labels(self.testcase, self.issue)
    self.assertIn('A', self.issue.components)
    self.assertIn('AA>B', self.issue.components)
    self.assertIn('D', self.issue.components)
    self.assertIn('Test-Predator-Auto-Components', self.issue.labels)


@test_utils.with_cloud_emulators('datastore')
class UpdateIssueCCsFromOwnersFileTest(unittest.TestCase):
  """Tests for update_issue_ccs_from_owners_file."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.is_oss_fuzz',
    ])
    helpers.patch_environ(self)

    self.issue = appengine_test_utils.create_generic_issue()
    self.testcase = test_utils.create_generic_testcase()

    # We'll generally want to assume we have an unassigned issue.
    self.issue.assignee = ''
    self.issue._monorail_issue.cc = []
    self.issue.status = 'Untriaged'

    self.testcase.set_metadata('issue_owners',
                               'dev1@example1.com,dev2@example2.com')
    self.mock.is_oss_fuzz.return_value = False
    self.policy = issue_tracker_policy.get('test-project')

  def test_skipped_issue_closed(self):
    """Test that we don't add ccs to closed issues."""
    self.issue.status = 'Fixed'
    self.issue._monorail_issue.open = False
    cleanup.update_issue_ccs_from_owners_file(self.policy, self.testcase,
                                              self.issue)
    self.assertEqual('', self.issue._monorail_issue.comment)
    six.assertCountEqual(self, [], self.issue.ccs)
    self.assertNotIn('ClusterFuzz-Auto-CC', self.issue.labels)

  def test_skipped_issue_updated_once(self):
    """Test that we don't add ccs if we added ccs once already."""
    comment = appengine_test_utils.create_generic_issue_comment(
        labels=['ClusterFuzz-Auto-CC'])
    self.assertEqual('', self.issue._monorail_issue.comment)
    self.issue._monorail_issue.comments.append(comment)
    cleanup.update_issue_ccs_from_owners_file(self.policy, self.testcase,
                                              self.issue)
    six.assertCountEqual(self, [], self.issue.ccs)

  def test_skipped_no_testcase_metadata(self):
    """Test that we don't add ccs if there are no issue_owners key in testcase
    metadata."""
    self.testcase.delete_metadata('issue_owners')
    cleanup.update_issue_ccs_from_owners_file(self.policy, self.testcase,
                                              self.issue)
    self.assertEqual('', self.issue._monorail_issue.comment)
    six.assertCountEqual(self, [], self.issue.ccs)
    self.assertNotIn('ClusterFuzz-Auto-CC', self.issue.labels)

  def test_skipped_empty_testcase_metadata(self):
    """Test that we don't add ccs if owners list is empty in testcase
    metadata."""
    self.testcase.set_metadata('issue_owners', '')
    cleanup.update_issue_ccs_from_owners_file(self.policy, self.testcase,
                                              self.issue)
    self.assertEqual('', self.issue._monorail_issue.comment)
    six.assertCountEqual(self, [], self.issue.ccs)
    self.assertNotIn('ClusterFuzz-Auto-CC', self.issue.labels)

  def test_skipped_ccs_already_added_and_metadata_set(self):
    """Test that we don't add ccs if ccs are added already and metadata has
    has_issue_ccs_from_owners_file attribute."""
    self.testcase.set_metadata('has_issue_ccs_from_owners_file', True)
    cleanup.update_issue_ccs_from_owners_file(self.policy, self.testcase,
                                              self.issue)
    self.assertEqual('', self.issue._monorail_issue.comment)
    six.assertCountEqual(self, [], self.issue.ccs)
    self.assertNotIn('ClusterFuzz-Auto-CC', self.issue.labels)

  def test_skipped_ccs_alread_added_and_metadata_set(self):
    """Test that we don't add ccs if ccs are added already."""
    self.issue.ccs.add('dev1@example1.com')
    self.issue.ccs.add('dev2@example2.com')
    cleanup.update_issue_ccs_from_owners_file(self.policy, self.testcase,
                                              self.issue)
    self.assertEqual(
        True, self.testcase.get_metadata('has_issue_ccs_from_owners_file'))
    self.assertEqual('', self.issue._monorail_issue.comment)
    six.assertCountEqual(self, ['dev1@example1.com', 'dev2@example2.com'],
                         sorted(self.issue.ccs))
    self.assertNotIn('ClusterFuzz-Auto-CC', self.issue.labels)

  def test_add_ccs_with_some_initial_ones(self):
    """Test that we only add new ccs if some are added already."""
    self.issue._monorail_issue.cc = ['dev1@example1.com']
    cleanup.update_issue_ccs_from_owners_file(self.policy, self.testcase,
                                              self.issue)
    self.assertEqual(
        'Automatically adding ccs based on OWNERS file / target commit history.'
        '\n\nIf this is incorrect, please add the ClusterFuzz-Wrong label.',
        self.issue._monorail_issue.comment)
    six.assertCountEqual(self, ['dev1@example1.com', 'dev2@example2.com'],
                         sorted(self.issue.ccs))
    self.assertIn('ClusterFuzz-Auto-CC', self.issue.labels)

  def test_add_ccs_without_any_initial_ones(self):
    """Test adding of ccs with none already existing on the issue."""
    self.mock.is_oss_fuzz.return_value = True
    cleanup.update_issue_ccs_from_owners_file(self.policy, self.testcase,
                                              self.issue)
    self.assertEqual(
        'Automatically adding ccs based on OWNERS file / target commit history.'
        '\n\nIf this is incorrect, '
        'please file a bug on https://github.com/google/oss-fuzz/issues/new.',
        self.issue._monorail_issue.comment)
    six.assertCountEqual(self, ['dev1@example1.com', 'dev2@example2.com'],
                         sorted(self.issue.ccs))
    self.assertIn('ClusterFuzz-Auto-CC', self.issue.labels)


@test_utils.with_cloud_emulators('datastore')
class UpdateIssueLabelsForFlakyTestcaseTest(unittest.TestCase):
  """Tests for update_issue_labels_for_flaky_testcase."""

  def setUp(self):
    self.issue = appengine_test_utils.create_generic_issue()
    self.testcase = test_utils.create_generic_testcase()
    self.policy = issue_tracker_policy.get('test-project')

  def test_mark_unreproducible_if_reproducible_change(self):
    """Test that we change label on issue if the testcase is now flaky."""
    self.issue.labels.add('Reproducible')
    self.testcase.one_time_crasher_flag = True
    self.testcase.put()
    cleanup.update_issue_labels_for_flaky_testcase(self.policy, self.testcase,
                                                   self.issue)

    self.assertNotIn('Reproducible', self.issue.labels)
    self.assertIn('Unreproducible', self.issue.labels)
    self.assertEqual(
        'ClusterFuzz testcase 1 appears to be flaky, '
        'updating reproducibility label.', self.issue._monorail_issue.comment)

  def test_skip_if_unreproducible(self):
    """Test that we don't change labels if the testcase is unreproducible and
    issue is already marked unreproducible."""
    self.issue.labels.add('Unreproducible')
    self.testcase.one_time_crasher_flag = True
    self.testcase.put()
    cleanup.update_issue_labels_for_flaky_testcase(self.policy, self.testcase,
                                                   self.issue)

    self.assertNotIn('Reproducible', self.issue.labels)
    self.assertIn('Unreproducible', self.issue.labels)
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_skip_if_reproducible(self):
    """Test that we don't change labels if the testcase is reproducible."""
    self.issue.labels.add('Reproducible')
    self.testcase.one_time_crasher_flag = False
    self.testcase.put()
    cleanup.update_issue_labels_for_flaky_testcase(self.policy, self.testcase,
                                                   self.issue)

    self.assertIn('Reproducible', self.issue.labels)
    self.assertNotIn('Unreproducible', self.issue.labels)
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_skip_if_another_reproducible_testcase(self):
    """Test that we don't change label on issue if another reproducible
    testcase exists."""
    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.one_time_crasher_flag = False
    similar_testcase.open = True
    similar_testcase.put()

    self.issue.labels.add('Reproducible')
    self.testcase.one_time_crasher_flag = True
    cleanup.update_issue_labels_for_flaky_testcase(self.policy, self.testcase,
                                                   self.issue)

    self.assertIn('Reproducible', self.issue.labels)
    self.assertNotIn('Unreproducible', self.issue.labels)
    self.assertEqual('', self.issue._monorail_issue.comment)


@test_utils.with_cloud_emulators('datastore')
class UpdateIssueOwnerAndCCsFromPredatorResultsTest(unittest.TestCase):
  """Tests for update_issue_owner_and_ccs_from_predator_results."""

  def setUp(self):
    self.issue = appengine_test_utils.create_generic_issue()
    self.testcase = test_utils.create_generic_testcase()

    # We'll generally want to assume we have an unassigned issue.
    self.issue.assignee = ''
    self.issue.status = 'Untriaged'
    self.policy = issue_tracker_policy.get('test-project')

    # Set the metadata to a generic result that would lead to an update,
    # assuming no other conditions are violated.
    self.testcase.set_metadata(
        'predator_result', {
            'result': {
                'suspected_cls': [{
                    'author': 'a@example.com',
                    'description': 'blah',
                    'url': 'url'
                },]
            }
        })

  def test_owner_assigned(self):
    """Ensure that we set the owner when appropriate."""
    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue)
    self.assertEqual(self.issue.assignee, 'a@example.com')
    self.assertEqual(self.issue.status, 'Assigned')
    self.assertIn('Test-Predator-Auto-Owner', self.issue.labels)

  def test_single_owner_cced_if_specified(self):
    """Ensure that we cc single authors if assignment is disabled."""
    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue, only_allow_ccs=True)
    self.assertEqual(self.issue.assignee, '')
    self.assertEqual(self.issue.status, 'Untriaged')
    self.assertIn('a@example.com', self.issue.ccs)
    self.assertIn('Test-Predator-Auto-CC', self.issue.labels)

  def test_closed_not_updated(self):
    """Ensure that we don't set owners for closed issues."""
    self.issue.status = 'Fixed'
    self.issue._monorail_issue.open = False

    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue)
    self.assertEqual(self.issue.assignee, '')
    self.assertEqual(self.issue.status, 'Fixed')
    self.assertNotIn('Test-Predator-Auto-Owner', self.issue.labels)

  def test_owner_not_reassigned(self):
    """Ensure that we don't overwrite already assigned owners."""
    self.issue.status = 'Assigned'
    self.issue.assignee = 'b@example.com'

    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue)
    self.assertEqual(self.issue.assignee, 'b@example.com')
    self.assertEqual(self.issue.status, 'Assigned')
    self.assertNotIn('Test-Predator-Auto-Owner', self.issue.labels)

  def test_skipped_if_already_updated(self):
    """Ensure that we don't try to update the same issue twice."""
    comment = appengine_test_utils.create_generic_issue_comment(
        labels=['Test-Predator-Auto-Owner'])
    self.issue._monorail_issue.comments.append(comment)

    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue)
    self.assertEqual(self.issue.assignee, '')
    self.assertEqual(self.issue.status, 'Untriaged')
    self.assertNotIn('Test-Predator-Auto-Owner', self.issue.labels)

  def test_skipped_if_previously_assigned(self):
    """Ensure that we don't assign to someone who was already the owner."""
    comment = appengine_test_utils.create_generic_issue_comment()
    comment.owner = 'a@example.com'
    self.issue._monorail_issue.comments.append(comment)

    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue)
    self.assertEqual(self.issue.assignee, '')
    self.assertEqual(self.issue.status, 'Untriaged')
    self.assertNotIn('Test-Predator-Auto-Owner', self.issue.labels)

  def test_skipped_if_no_cls(self):
    """Ensure that we do nothing if we have no suspected CLs."""
    self.testcase.set_metadata('predator_result',
                               {'result': {
                                   'suspected_cls': []
                               }})

    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue)
    self.assertEqual(self.issue.assignee, '')
    self.assertEqual(self.issue.status, 'Untriaged')
    self.assertNotIn('Test-Predator-Auto-Owner', self.issue.labels)

  def test_add_ccs_if_multiple_cls(self):
    """Ensure that we only cc when we have multiple suspected CLs."""
    self.testcase.set_metadata(
        'predator_result', {
            'result': {
                'suspected_cls': [
                    {
                        'author': 'a@example.com',
                        'description': 'blah',
                        'url': 'url'
                    },
                    {
                        'author': 'b@example.com',
                        'description': 'halb',
                        'url': 'lru'
                    },
                ]
            }
        })

    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue)
    self.assertEqual(self.issue.assignee, '')
    self.assertEqual(self.issue.status, 'Untriaged')
    self.assertNotIn('Test-Predator-Auto-Owner', self.issue.labels)
    self.assertIn('Test-Predator-Auto-CC', self.issue.labels)
    self.assertIn('a@example.com', self.issue.ccs)
    self.assertIn('b@example.com', self.issue.ccs)

  def test_skipped_if_previously_cced_and_metadata_set(self):
    """Ensure that we don't re-cc authors who were cced in the past and have
    has_issue_ccs_from_predator_results set in metadata."""
    self.testcase.set_metadata('has_issue_ccs_from_predator_results', True)
    self.testcase.set_metadata(
        'predator_result', {
            'result': {
                'suspected_cls': [
                    {
                        'author': 'a@example.com',
                        'description': 'blah',
                        'url': 'url'
                    },
                    {
                        'author': 'b@example.com',
                        'description': 'halb',
                        'url': 'lru'
                    },
                ]
            }
        })

    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue, only_allow_ccs=True)
    self.assertNotIn('a@example.com', self.issue.ccs)
    self.assertNotIn('b@example.com', self.issue.ccs)
    self.assertNotIn('Test-Predator-Auto-CC', self.issue.labels)

  def test_skipped_if_previously_cced_and_metadata_not_set(self):
    """Ensure that we don't re-cc authors who were cced in the past."""
    comment = appengine_test_utils.create_generic_issue_comment()
    comment.cc = ['-a@example.com']
    self.issue._monorail_issue.comments.append(comment)

    self.testcase.set_metadata(
        'predator_result', {
            'result': {
                'suspected_cls': [
                    {
                        'author': 'a@example.com',
                        'description': 'blah',
                        'url': 'url'
                    },
                    {
                        'author': 'b@example.com',
                        'description': 'halb',
                        'url': 'lru'
                    },
                ]
            }
        })

    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue)
    self.assertNotIn('a@example.com', self.issue.ccs)
    self.assertIn('b@example.com', self.issue.ccs)
    self.assertIn('Test-Predator-Auto-CC', self.issue.labels)

  def test_skipped_if_malformed_cl(self):
    """Ensure that we do nothing if the suspected CL is malformed."""
    self.testcase.set_metadata('predator_result',
                               {'result': {
                                   'suspected_cls': [{
                                       'url': 'url'
                                   },]
                               }})

    cleanup.update_issue_owner_and_ccs_from_predator_results(
        self.policy, self.testcase, self.issue)
    self.assertEqual(self.issue.assignee, '')
    self.assertEqual(self.issue.status, 'Untriaged')
    self.assertNotIn('Test-Predator-Auto-Owner', self.issue.labels)


@test_utils.with_cloud_emulators('datastore')
class NotifyIssueIfTestcaseIsInvalidTest(unittest.TestCase):
  """Tests for notify_issue_if_testcase_is_invalid."""

  def setUp(self):
    self.issue = appengine_test_utils.create_generic_issue()
    self.testcase = test_utils.create_generic_testcase()

    # Assume a test case associated with an assigned issue.
    self.issue.status = 'Assigned'
    self.testcase.bug_information = '123456'
    self.policy = issue_tracker_policy.get('test-project')

  def test_skipped_if_no_issue(self):
    """Ensure that we handle the case where there is no issue."""
    self.testcase.bug_information = None

    # Simply ensure that we don't throw an exception in this case.
    cleanup.notify_issue_if_testcase_is_invalid(self.policy, self.testcase,
                                                None)

  def test_skipped_if_closed_issue(self):
    """Ensure that we ignore issues that are already closed."""
    self.issue.status = 'Fixed'
    self.issue._monorail_issue.open = False
    cleanup.notify_issue_if_testcase_is_invalid(self.policy, self.testcase,
                                                self.issue)
    self.assertEqual(self.issue._monorail_issue.comment, '')

  def test_skipped_if_unmarked_issue(self):
    """Ensure that we ignore issues that have valid fuzzers."""
    cleanup.notify_issue_if_testcase_is_invalid(self.policy, self.testcase,
                                                self.issue)
    self.assertEqual(self.issue._monorail_issue.comment, '')

  def test_notified_if_fuzzer_was_deleted(self):
    """Ensure that we comment on issues that have invalid fuzzers."""
    self.testcase.set_metadata('fuzzer_was_deleted', True)
    cleanup.notify_issue_if_testcase_is_invalid(self.policy, self.testcase,
                                                self.issue)
    self.assertIn('is associated with an obsolete fuzzer',
                  self.issue._monorail_issue.comment)
    self.assertIn(ISSUE_INVALID_FUZZER_LABEL, self.issue.labels)

  def test_not_notified_if_fuzzer_was_deleted_and_notified(self):
    """Ensure that we don't comment again on issues that have invalid fuzzers
    and we have commented once."""
    self.testcase.set_metadata('fuzzer_was_deleted', True)
    self.issue._monorail_issue.comments += [
        appengine_test_utils.create_generic_issue_comment(
            labels=[ISSUE_INVALID_FUZZER_LABEL])
    ]
    cleanup.notify_issue_if_testcase_is_invalid(self.policy, self.testcase,
                                                self.issue)
    self.assertNotIn('is associated with an obsolete fuzzer',
                     self.issue._monorail_issue.comment)


@test_utils.with_cloud_emulators('datastore')
class NotifyUploaderIfTestcaseIsProcessed(unittest.TestCase):
  """Tests for notify_uploader_when_testcase_is_processed."""

  def setUp(self):
    helpers.patch(self, [
        'handlers.cron.cleanup._update_issue_security_severity_and_get_comment',
        'libs.issue_management.issue_filer.update_issue_impact_labels',
        'libs.mail.send',
    ])

    self.issue = appengine_test_utils.create_generic_issue()
    self.testcase = test_utils.create_generic_testcase()
    self.testcase_id = self.testcase.key.id()
    self.uploader_email = 'uploader@email.com'
    self.policy = issue_tracker_policy.get('test-project')

    data_types.Config(url='url', reproduction_help_url='repro_help_url').put()

  def _get_notification(self):
    """Return notification entity for our testcase."""
    return data_types.Notification.query(
        data_types.Notification.testcase_id == self.testcase_id).get()

  def test_no_upload_metadata(self):
    """Ensure that we don't send notification if there is no upload metadata."""
    cleanup.notify_uploader_when_testcase_is_processed(
        self.policy, self.testcase, self.issue)

    self.assertEqual(0, self.mock.send.call_count)
    self.assertIsNone(self._get_notification())

  def test_upload_metadata_with_no_uploader_email(self):
    """Ensure that we don't send notification if there is no uploader email."""
    data_types.TestcaseUploadMetadata(
        testcase_id=self.testcase_id, uploader_email=None, bundled=False).put()
    cleanup.notify_uploader_when_testcase_is_processed(
        self.policy, self.testcase, self.issue)

    self.assertEqual(0, self.mock.send.call_count)
    self.assertIsNone(self._get_notification())

  def test_upload_metadata_with_multiple_testcases(self):
    """Ensure that we don't send notification if this a bundled metadata archive
    (with multiple testcases)."""
    data_types.TestcaseUploadMetadata(
        testcase_id=self.testcase_id,
        uploader_email=self.uploader_email,
        bundled=False).put()
    cleanup.notify_uploader_when_testcase_is_processed(
        self.policy, self.testcase, self.issue)

    self.assertEqual(0, self.mock.send.call_count)
    self.assertIsNone(self._get_notification())

  def test_critical_tasks_not_completed(self):
    """Ensure that we don't send notification if critical tasks not complete."""
    data_types.TestcaseUploadMetadata(
        testcase_id=self.testcase_id,
        uploader_email=self.uploader_email,
        bundled=False).put()
    self.testcase.minimized_keys = None
    self.testcase.regression = None
    self.testcase.put()
    cleanup.notify_uploader_when_testcase_is_processed(
        self.policy, self.testcase, self.issue)

    self.assertEqual(0, self.mock.send.call_count)

  def test_pending_testcase(self):
    """Ensure that notification is not sent with a pending testcase."""
    data_types.TestcaseUploadMetadata(
        testcase_id=self.testcase_id,
        uploader_email=self.uploader_email,
        bundled=False).put()
    self.testcase.status = 'Pending'
    self.testcase.one_time_crasher_flag = False
    self.testcase.put()

    cleanup.notify_uploader_when_testcase_is_processed(
        self.policy, self.testcase, self.issue)

    self.assertEqual(0, self.mock.send.call_count)

  def test_notification_sent_with_regular_testcase(self):
    """Ensure that notification is sent with a regular testcase."""
    data_types.TestcaseUploadMetadata(
        testcase_id=self.testcase_id,
        uploader_email=self.uploader_email,
        bundled=False).put()
    self.testcase.status = 'Processed'
    self.testcase.minimized_keys = 'some-key'
    self.testcase.regression = '1:2'
    self.testcase.is_impact_set_flag = True
    self.testcase.put()

    cleanup.notify_uploader_when_testcase_is_processed(
        self.policy, self.testcase, self.issue)

    self.mock.send.assert_called_once_with(
        'uploader@email.com', 'Your testcase upload 1 analysis is complete.',
        'Detailed Report: https://test-clusterfuzz.appspot.com/'
        'testcase?key=1<br><br>'
        'Fuzzer: fuzzer1<br>'
        'Job Type: test_content_shell_drt<br>'
        'Crash Type: fake type<br>'
        'Crash Address: 0xdeadbeef<br>'
        'Crash State:<br>'
        '  ...see report...<br>'
        'Sanitizer: address (ASAN)<br><br>'
        'Regressed: https://test-clusterfuzz.appspot.com/revisions?'
        'job=test_content_shell_drt&range=1:2<br><br>'
        'Reproducer Testcase: '
        'https://test-clusterfuzz.appspot.com/download?testcase_id=1<br><br>'
        'See repro_help_url for instructions to reproduce this bug locally.'
        '<br><br>'
        'If you suspect that the result above is incorrect, '
        'try re-doing that job on the testcase report page.')
    self.assertIsNotNone(self._get_notification())

  def test_notification_sent_with_unreproducible_testcase(self):
    """Ensure that notification is sent with an unreproducible testcase."""
    data_types.TestcaseUploadMetadata(
        testcase_id=self.testcase_id,
        uploader_email=self.uploader_email,
        bundled=False).put()
    self.testcase.status = 'Unreproducible'
    self.testcase.one_time_crasher_flag = False
    self.testcase.put()

    cleanup.notify_uploader_when_testcase_is_processed(
        self.policy, self.testcase, self.issue)

    self.mock.send.assert_called_once_with(
        'uploader@email.com', 'Your testcase upload 1 analysis is complete.',
        'Testcase 1 failed to reproduce the crash. '
        'Please inspect the program output at '
        'https://test-clusterfuzz.appspot.com/testcase?key=1.<br><br>'
        'If you suspect that the result above is incorrect, '
        'try re-doing that job on the testcase report page.')
    self.assertIsNotNone(self._get_notification())

  def test_notification_sent_with_one_issue_update_when_quiet_flag_set(self):
    """Ensure that notification is sent and issue is updated only once when
    quiet flag is set."""
    data_types.TestcaseUploadMetadata(
        testcase_id=self.testcase_id,
        uploader_email=self.uploader_email,
        bundled=False,
        bug_information=str(self.issue.id),
        quiet_flag=True).put()
    self.testcase.status = 'Processed'
    self.testcase.minimized_keys = 'some-key'
    self.testcase.regression = '1:2'
    self.testcase.is_impact_set_flag = True
    self.testcase.put()

    for _ in range(3):
      cleanup.notify_uploader_when_testcase_is_processed(
          self.policy, self.testcase, self.issue)

    self.assertEqual(
        1, self.mock._update_issue_security_severity_and_get_comment.call_count)
    self.assertIsNotNone(self._get_notification())


@test_utils.with_cloud_emulators('datastore')
class CleanupUnusedFuzzTargetsTest(unittest.TestCase):
  """Tests for cleanup_unused_fuzz_targets_and_jobs."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
    ])
    self.mock.utcnow.return_value = datetime.datetime(2018, 1, 31)

  def test_cleanup_unused_fuzz_targets_and_jobs(self):
    """Test cleaning up fuzz targets."""
    # FuzzTarget should be removed. All FuzzTargetJobs are older than the
    # threshold.
    data_types.FuzzTarget(
        engine='libFuzzer', binary='binary1', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_binary1',
        job='job1',
        last_run=datetime.datetime(2018, 1, 15)).put()

    # FuzzTarget should be removed. No FuzzTargetJobs.
    data_types.FuzzTarget(
        engine='libFuzzer', binary='binary2', project='test-project').put()

    # FuzzTarget should not be removed. Has 1 FuzzTargetJob left after removing
    # old ones.
    data_types.FuzzTarget(
        engine='libFuzzer', binary='binary3', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_binary3',
        job='job1',
        last_run=datetime.datetime(2018, 1, 20)).put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_binary3',
        job='job2',
        last_run=datetime.datetime(2018, 1, 15)).put()

    # FuzzTarget should not be removed. All FuzzTargetJob valid.
    data_types.FuzzTarget(
        engine='libFuzzer', binary='binary4', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_binary4',
        job='job1',
        last_run=datetime.datetime(2018, 1, 20)).put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_binary4',
        job='job2',
        last_run=datetime.datetime(2018, 1, 20)).put()

    cleanup.cleanup_unused_fuzz_targets_and_jobs()

    six.assertCountEqual(
        self, ['libFuzzer_binary3', 'libFuzzer_binary4'],
        list([t.key.id() for t in data_types.FuzzTarget.query()]))
    six.assertCountEqual(
        self, [
            'libFuzzer_binary3/job1', 'libFuzzer_binary4/job1',
            'libFuzzer_binary4/job2'
        ], list([t.key.id() for t in data_types.FuzzTargetJob.query()]))


@test_utils.with_cloud_emulators('datastore')
class CleanupUnusedHeartbeatsTest(unittest.TestCase):
  """Tests for cleaning up heartbeat entities."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
    ])

    self.mock.utcnow.return_value = datetime.datetime(2018, 1, 31)

  def test_cleanup(self):
    """Test cleanup_unused_heartbeats."""
    data_types.Heartbeat(last_beat_time=datetime.datetime(2018, 1, 14)).put()
    data_types.Heartbeat(last_beat_time=datetime.datetime(2018, 1, 15)).put()
    data_types.Heartbeat(last_beat_time=datetime.datetime(2018, 1, 16)).put()
    data_types.Heartbeat(last_beat_time=datetime.datetime(2018, 1, 17)).put()
    data_types.Heartbeat(last_beat_time=datetime.datetime(2018, 2, 1)).put()
    cleanup.cleanup_unused_heartbeats()

    six.assertCountEqual(self, [
        {
            'task_payload': None,
            'source_version': None,
            'task_end_time': None,
            'last_beat_time': datetime.datetime(2018, 1, 16, 0, 0),
            'bot_name': None,
            'platform_id': None,
            'keywords': [],
        },
        {
            'task_payload': None,
            'source_version': None,
            'task_end_time': None,
            'last_beat_time': datetime.datetime(2018, 1, 17, 0, 0),
            'bot_name': None,
            'platform_id': None,
            'keywords': [],
        },
        {
            'task_payload': None,
            'source_version': None,
            'task_end_time': None,
            'last_beat_time': datetime.datetime(2018, 2, 1, 0, 0),
            'bot_name': None,
            'platform_id': None,
            'keywords': [],
        },
    ], [e.to_dict() for e in data_types.Heartbeat.query()])


class UpdateSeverityLabelsTest(unittest.TestCase):
  """Tests for updating severity labels."""

  def setUp(self):
    self.testcase = data_types.Testcase()
    self.issue = appengine_test_utils.create_generic_issue()
    self.policy = issue_tracker_policy.get('test-project')

  def test_add_missing_severity(self):
    """Test updating missing severity."""
    self.testcase.security_severity = data_types.SecuritySeverity.HIGH
    result = cleanup._update_issue_security_severity_and_get_comment(
        self.policy, self.testcase, self.issue)
    self.assertIn('Security_Severity-High', self.issue.labels)
    self.assertIn('A recommended severity was added to this bug.', result)

  def test_add_same_severity(self):
    """Test correct severity already set."""
    self.testcase.security_severity = data_types.SecuritySeverity.HIGH
    self.issue.labels.add('Security_severity-High')
    result = cleanup._update_issue_security_severity_and_get_comment(
        self.policy, self.testcase, self.issue)
    self.assertIn('Security_Severity-High', self.issue.labels)
    self.assertEqual('', result)

  def test_add_different_severity(self):
    """Test incorrect severity set."""
    self.testcase.security_severity = data_types.SecuritySeverity.HIGH
    self.issue.labels.add('Security_Severity-Medium')
    result = cleanup._update_issue_security_severity_and_get_comment(
        self.policy, self.testcase, self.issue)
    self.assertNotIn('Security_Severity-High', self.issue.labels)
    self.assertIn('Security_Severity-Medium', self.issue.labels)
    self.assertIn('different from what was assigned', result)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""batcher tests."""
import datetime
import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import test_utils
from local.butler.scripts import batcher


@test_utils.with_cloud_emulators('datastore')
class BatcherTest(unittest.TestCase):
  """Test batcher."""

  def setUp(self):
    self.testcase_ids = []
    for i in range(100):
      testcase = data_types.Testcase()
      testcase.timestamp = datetime.datetime.fromtimestamp(i)
      testcase.put()
      self.testcase_ids.append(testcase.key.id())

  def _test(self, batch_size, expected_batch_count, expected_count):
    """Test when the limit is too large."""
    query = data_types.Testcase.query().order(data_types.Testcase.timestamp)
    list_of_testcases = list(batcher.iterate(query, batch_size=batch_size))

    self.assertEqual(expected_batch_count, len(list_of_testcases))
    count = 0
    for testcases in list_of_testcases:
      for testcase in testcases:
        self.assertEqual(self.testcase_ids[count], testcase.key.id())
        count += 1
    self.assertEqual(expected_count, count)

  def test_batch(self):
    """Test batching."""
    self._test(2, 50, 100)

  def test_batch_non_multiple(self):
    """Test when the batch size is not a multiple."""
    self._test(7, 15, 100)

  def test_too_large_batch(self):
    """Test when the batch is too large."""
    self._test(105, 1, 100)

  def test_exact_batch(self):
    """Test when the batch is exactly the number of items."""
    self._test(100, 1, 100)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for sync_admins."""

import unittest

import flask
import six
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import sync_admins


@test_utils.with_cloud_emulators('datastore')
class SyncAdminsTest(unittest.TestCase):
  """Tests for sync_admins."""

  def setUp(self):
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/sync-admins', view_func=sync_admins.Handler.as_view('/sync-admins'))
    self.app = webtest.TestApp(flaskapp)

    test_helpers.patch(self, [
        'googleapiclient.discovery.build',
        'handlers.base_handler.Handler.is_cron',
    ])

    data_types.Admin(id='remove@email.com', email='remove@email.com').put()
    data_types.Admin(id='user1@email.com', email='keep@email.com').put()

    self.mock.is_cron.return_value = True
    self.mock.build(
        serviceName='',
        version='').projects().getIamPolicy().execute.return_value = {
            'version':
                1,
            'etag':
                'etag',
            'bindings': [
                {
                    'role':
                        'roles/appengine.appAdmin',
                    'members': [
                        'user:user1@email.com',
                        'serviceAccount:service_account@gserviceaccount.com',
                    ]
                },
                {
                    'role': 'roles/appengine.deployer',
                    'members': ['user:not_added@email.com',]
                },
                {
                    'role': 'roles/appengine.serviceAdmin',
                    'members': ['user:not_added@email.com',]
                },
                {
                    'role': 'roles/editor',
                    'members': [
                        'user:user2@email.com',
                        'user:user3@email.com',
                    ]
                },
                {
                    'role': 'roles/owner',
                    'members': ['user:user4@email.com',]
                },
                {
                    'role': 'roles/viewer',
                    'members': ['user:user5@email.com',]
                },
            ]
        }

  def test_sync_admins(self):
    """Test syncing admins."""
    self.app.get('/sync-admins')
    admins = data_types.Admin.query()
    six.assertCountEqual(self, [
        'user1@email.com',
        'user2@email.com',
        'user3@email.com',
        'user4@email.com',
        'user5@email.com',
    ], [admin.email for admin in admins])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the stack symbolizer module."""

import unittest

from clusterfuzz._internal.crash_analysis.stack_parsing import stack_symbolizer


class ChromeDsymHintsTests(unittest.TestCase):
  """Tests chrome_dsym_hints."""

  def test_standalone_executable(self):
    """Tests that standalone executable work as expected."""
    self.assertEqual([], stack_symbolizer.chrome_dsym_hints('/build/binary'))

  def test_standalone_framework_or_app(self):
    """Tests that standalone frame or app bundle work as expected."""
    self.assertEqual(
        ['/build/Content Shell.dSYM'],
        stack_symbolizer.chrome_dsym_hints('/build/Content Shell.app'))
    self.assertEqual(
        ['/build/Content Shell.dSYM'],
        stack_symbolizer.chrome_dsym_hints('/build/Content Shell.framework'))

  def test_nested_bundles(self):
    """Tests that two or three nested bundles work as expected."""
    self.assertEqual(
        ['/build/Content Shell Helper.dSYM'],
        stack_symbolizer.chrome_dsym_hints(
            '/build/Content Shell.app/Contents'
            '/Frameworks/Content Shell Framework.framework/Versions/C/Helpers'
            '/Content Shell Helper.app/Contents/MacOS/Content Shell Helper'))
    self.assertEqual(
        ['/build/Content Shell Helper.dSYM'],
        stack_symbolizer.chrome_dsym_hints(
            '/build/Content Shell.app/Contents'
            '/Versions/C/Helpers'
            '/Content Shell Helper.app/Contents/MacOS/Content Shell Helper'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gestures handler."""

from clusterfuzz._internal.platforms import android
from clusterfuzz._internal.platforms import linux
from clusterfuzz._internal.platforms import windows
from clusterfuzz._internal.system import environment


def get_gestures(gesture_count):
  """Return a list of random gestures."""
  plt = environment.platform()

  if environment.is_android(plt):
    return android.gestures.get_random_gestures(gesture_count)
  if plt == 'LINUX':
    return linux.gestures.get_random_gestures(gesture_count)
  if plt == 'WINDOWS':
    return windows.gestures.get_random_gestures(gesture_count)

  return []



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for jobs."""
import collections
import string
import unittest
from unittest import mock

import flask
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers import jobs
from libs import form


@test_utils.with_cloud_emulators('datastore')
class JobsTest(unittest.TestCase):
  """Jobs tests."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.access.has_access',
        'libs.access.get_access',
        'libs.helpers.get_user_email',
        'libs.gcs.prepare_blob_upload',
    ])
    self.mock.prepare_blob_upload.return_value = (
        collections.namedtuple('GcsUpload', [])())
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=jobs.JsonHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

  def _create_job(self,
                  name,
                  environment_string,
                  description='',
                  platform='LINUX'):
    """Create a test job."""
    job = data_types.Job()
    job.name = name
    if environment_string.strip():
      job.environment_string = environment_string
    job.platform = platform
    job.descripton = description
    job.put()

    return job

  def test_post(self):
    """Test post method."""
    self.mock.has_access.return_value = True
    expected_items = {1: [], 2: [], 3: []}

    for job_num, job_suffix in enumerate(string.ascii_lowercase):
      job_name = "test_job_" + job_suffix
      job = self._create_job(job_name, 'APP_NAME = launcher.py\n')
      expected_items[(job_num // 10) + 1].append(job.key.id())

    resp = self.app.post_json('/', {'page': 1})
    self.assertListEqual(expected_items[1],
                         [item['id'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'page': 2})
    self.assertListEqual(expected_items[2],
                         [item['id'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'page': 3})
    self.assertListEqual(expected_items[3],
                         [item['id'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'page': 4})
    self.assertListEqual([], [item['id'] for item in resp.json['items']])

  def test_fuzzers_result(self):
    """Test fuzzers result obtained in post method."""
    self.mock.has_access.return_value = True

    job = self._create_job('test_job', 'APP_NAME = launcher.py\n')
    fuzzer = data_types.Fuzzer()
    fuzzer.name = 'fuzzer'
    fuzzer.jobs = ['test_job']
    fuzzer.put()

    resp = self.app.post_json('/', {'page': 1})
    self.assertListEqual([job.key.id()],
                         [item['id'] for item in resp.json['items']])
    self.assertIn('fuzzer', resp.json['items'][0]['fuzzers'])


@test_utils.with_cloud_emulators('datastore')
class JobsSearchTest(unittest.TestCase):
  """Jobs search tests."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.access.has_access',
        'libs.access.get_access',
        'libs.helpers.get_user_email',
        'libs.gcs.prepare_blob_upload',
    ])
    self.mock.prepare_blob_upload.return_value = (
        collections.namedtuple('GcsUpload', [])())
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=jobs.JsonHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

  def _create_job(self,
                  name,
                  environment_string,
                  description='',
                  platform='LINUX'):
    """Create a test job."""
    job = data_types.Job()
    job.name = name
    if environment_string.strip():
      job.environment_string = environment_string
    job.platform = platform
    job.descripton = description
    job.put()

    return job

  def test_post(self):
    """Test post method."""
    self.mock.has_access.return_value = True

    job_asan = self._create_job('test_job_asan', 'PROJECT_NAME = proj_1\n')
    job_ubsan = self._create_job('test_job_ubsan', 'PROJECT_NAME = proj_2\n')

    resp = self.app.post_json('/', {'q': "asan"})
    self.assertListEqual([job_asan.key.id()],
                         [item['id'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': "ubsan"})
    self.assertListEqual([job_ubsan.key.id()],
                         [item['id'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': "testing"})
    self.assertListEqual([], [item['id'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': "test"})
    self.assertListEqual(
        [job_asan.key.id(), job_ubsan.key.id()],
        [item['id'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': "1"})
    self.assertListEqual([job_asan.key.id()],
                         [item['id'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': "2"})
    self.assertListEqual([job_ubsan.key.id()],
                         [item['id'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': "proj"})
    self.assertListEqual(
        [job_asan.key.id(), job_ubsan.key.id()],
        [item['id'] for item in resp.json['items']])


@test_utils.with_cloud_emulators('datastore')
class JobsUpdateTest(unittest.TestCase):
  """Job update tests."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.auth.get_current_user',
        'libs.access.has_access',
        'libs.access.get_access',
        'libs.gcs.prepare_blob_upload',
        'clusterfuzz._internal.fuzzing.fuzzer_selection.update_mappings_for_job',
    ])
    self.mock.get_current_user().email = 'test@user.com'
    self.mock.prepare_blob_upload.return_value = (
        collections.namedtuple('GcsUpload', [])())
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=jobs.UpdateJob.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

  def _create_job(self,
                  name,
                  environment_string,
                  description='',
                  platform='LINUX'):
    """Create a test job."""
    job = data_types.Job()
    job.name = name
    if environment_string.strip():
      job.environment_string = environment_string
    job.platform = platform
    job.descripton = description
    job.put()

    return job

  def test_post(self):
    """Test post method."""
    self.mock.has_access.return_value = True
    job = self._create_job('test_job', 'PROJECT_NAME = proj\n')
    resp = self.app.post(
        '/', {
            'csrf_token': form.generate_csrf_token(),
            'name': job.name,
            'desciption': job.description,
            'platform': job.platform,
            'fuzzers': ['test_fuzzer']
        },
        expect_errors=True)
    self.assertEqual(200, resp.status_int)
    self.mock.update_mappings_for_job.assert_called_with(
        mock.ANY, ['test_fuzzer'])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fuzzer_selection.py."""

import os
import unittest

import parameterized

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.fuzzing import fuzzer_selection
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


def _get_job_list_for_fuzzer(fuzzer):
  """Helper function to return the mappings for a fuzzer as a list."""
  query = data_types.FuzzerJob.query()
  query.filter(data_types.FuzzerJob.fuzzer == fuzzer.name)
  return [m.job for m in ndb_utils.get_all_from_query(query)]


def _get_fuzzer_list_for_job(job):
  """Helper function to return the mappings for a job as a list."""
  fuzzers = data_types.Fuzzer.query().filter(data_types.Fuzzer.jobs == job.name)
  return [fuzzer.name for fuzzer in fuzzers]


@test_utils.with_cloud_emulators('datastore')
class UpdateMappingsForFuzzerTest(unittest.TestCase):
  """Tests for the update_mappings_for_fuzzer function."""

  def setUp(self):
    job_1 = data_types.Job()
    job_1.name = 'job_1'
    job_1.environment_string = 'test = 1'
    job_1.put()

    job_2 = data_types.Job()
    job_2.name = 'job_2'
    job_2.environment_string = 'test = 2'
    job_2.put()

  def test_no_mappings(self):
    """Ensure that we do nothing if a fuzzer has no mappings."""
    fuzzer = data_types.Fuzzer()
    fuzzer.name = 'no_mappings'
    fuzzer.put()

    fuzzer_selection.update_mappings_for_fuzzer(fuzzer)

    self.assertEqual(_get_job_list_for_fuzzer(fuzzer), [])

  def test_new_addition(self):
    """Ensure that we add mappings for a new fuzzer."""
    fuzzer = data_types.Fuzzer()
    fuzzer.name = 'new_addition'
    fuzzer.jobs = ['job_1', 'job_2']
    fuzzer.put()

    fuzzer_selection.update_mappings_for_fuzzer(fuzzer)

    mappings = _get_job_list_for_fuzzer(fuzzer)
    self.assertIn('job_1', mappings)
    self.assertIn('job_2', mappings)

  def test_mapping_added(self):
    """Ensure that we properly add mappings for existing fuzzers."""
    fuzzer = data_types.Fuzzer()
    fuzzer.name = 'adding_jobs'
    fuzzer.jobs = ['job_1']
    fuzzer.put()

    fuzzer_selection.update_mappings_for_fuzzer(fuzzer)

    mappings = _get_job_list_for_fuzzer(fuzzer)
    self.assertIn('job_1', mappings)
    self.assertNotIn('job_2', mappings)

    fuzzer.jobs += ['job_2']
    fuzzer_selection.update_mappings_for_fuzzer(fuzzer)

    mappings = _get_job_list_for_fuzzer(fuzzer)
    self.assertIn('job_1', mappings)
    self.assertIn('job_2', mappings)

  def test_mapping_subsituted(self):
    """Ensure that mappings are subsituted properly."""
    fuzzer = data_types.Fuzzer()
    fuzzer.name = 'adding_jobs'
    fuzzer.jobs = ['job_1']
    fuzzer.put()

    fuzzer_selection.update_mappings_for_fuzzer(fuzzer)

    mappings = _get_job_list_for_fuzzer(fuzzer)
    self.assertIn('job_1', mappings)
    self.assertNotIn('job_2', mappings)

    fuzzer.jobs = ['job_2']
    fuzzer_selection.update_mappings_for_fuzzer(fuzzer)

    mappings = _get_job_list_for_fuzzer(fuzzer)
    self.assertNotIn('job_1', mappings)
    self.assertIn('job_2', mappings)

  def test_mapping_removed(self):
    """Ensure that mappings are removed properly."""
    fuzzer = data_types.Fuzzer()
    fuzzer.name = 'adding_jobs'
    fuzzer.jobs = ['job_1']
    fuzzer.put()

    fuzzer_selection.update_mappings_for_fuzzer(fuzzer, [])

    mappings = _get_job_list_for_fuzzer(fuzzer)
    self.assertEqual([], mappings)


@test_utils.with_cloud_emulators('datastore')
class UpdateMappingsForJobTest(unittest.TestCase):
  """Tests for the update_mappings_for_job function."""

  def setUp(self):
    self.fuzzer_1 = data_types.Fuzzer()
    self.fuzzer_1.name = 'fuzzer_1'
    self.fuzzer_1.put()

    self.fuzzer_2 = data_types.Fuzzer()
    self.fuzzer_2.name = 'fuzzer_2'
    self.fuzzer_2.put()

  def test_no_mappings(self):
    """Ensure that we do nothing if a job has no mappings."""
    job = data_types.Job()
    job.name = 'no_mappings'
    job.put()

    fuzzer_selection.update_mappings_for_job(job, [])

    self.assertEqual(_get_fuzzer_list_for_job(job), [])

  def test_new_addition(self):
    """Ensure that we add mappings for a new job."""
    job = data_types.Job()
    job.name = 'new_addition'
    job.put()

    fuzzer_selection.update_mappings_for_job(job, ['fuzzer_1', 'fuzzer_2'])

    mappings = _get_fuzzer_list_for_job(job)
    self.assertIn('fuzzer_1', mappings)
    self.assertIn('fuzzer_2', mappings)

  def test_mapping_added(self):
    """Ensure that we properly add mappings for existing jobs."""
    job = data_types.Job()
    job.name = 'adding_fuzzers'
    job.put()

    fuzzer_selection.update_mappings_for_job(job, ['fuzzer_1'])

    mappings = _get_fuzzer_list_for_job(job)
    self.assertIn('fuzzer_1', mappings)
    self.assertNotIn('fuzzer_2', mappings)

    fuzzer_selection.update_mappings_for_job(job, ['fuzzer_1', 'fuzzer_2'])

    mappings = _get_fuzzer_list_for_job(job)
    self.assertIn('fuzzer_1', mappings)
    self.assertIn('fuzzer_2', mappings)

  def test_mapping_substituted(self):
    """Ensure that mappings are substituted properly."""
    job = data_types.Job()
    job.name = 'substitute_fuzzers'
    job.put()

    fuzzer_selection.update_mappings_for_job(job, ['fuzzer_1'])

    mappings = _get_fuzzer_list_for_job(job)
    self.assertIn('fuzzer_1', mappings)
    self.assertNotIn('fuzzer_2', mappings)

    fuzzer_selection.update_mappings_for_job(job, ['fuzzer_2'])

    mappings = _get_fuzzer_list_for_job(job)
    self.assertNotIn('fuzzer_1', mappings)
    self.assertIn('fuzzer_2', mappings)

  def test_mapping_removed(self):
    """Ensure that mappings are removed properly."""
    job = data_types.Job()
    job.name = 'remove_fuzzer'
    job.put()

    self.fuzzer_1.jobs.append('remove_fuzzer')
    self.fuzzer_1.put()

    fuzzer_selection.update_mappings_for_job(job, [])

    mappings = _get_fuzzer_list_for_job(job)
    self.assertEqual([], mappings)


def _mock_random_weighted_choice(items, weight_attribute='weight'):  # pylint: disable=unused-argument
  """Mock random_weighted_choice."""
  # Always select the first element rather than a random one for the sake of
  # determinism.
  return items[0]


@test_utils.with_cloud_emulators('datastore')
class GetFuzzTaskPayloadTest(unittest.TestCase):
  """Tests for the get_fuzz_task_payload function."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.random_weighted_choice',
    ])

    self.mock.random_weighted_choice.side_effect = _mock_random_weighted_choice

  def test_no_mappings(self):
    """Ensure that we raise an exception if we don't find a task."""
    self.assertEqual(
        (None, None), fuzzer_selection.get_fuzz_task_payload(platform='linux'))

  @parameterized.parameterized.expand([('False',), ('True',)])
  def test_platform_restriction(self, local_development):
    """Ensure that we can find a task with a valid platform."""
    os.environ['LOCAL_DEVELOPMENT'] = local_development
    windows_mapping = data_types.FuzzerJob()
    windows_mapping.fuzzer = 'wrong_fuzzer'
    windows_mapping.job = 'job_1'
    windows_mapping.platform = 'windows'
    windows_mapping.put()

    data_types.FuzzerJobs(
        platform='windows', fuzzer_jobs=[windows_mapping]).put()

    self.assertEqual(
        (None, None), fuzzer_selection.get_fuzz_task_payload(platform='linux'))

    linux_mapping = data_types.FuzzerJob()
    linux_mapping.fuzzer = 'right_fuzzer'
    linux_mapping.job = 'job_2'
    linux_mapping.platform = 'linux'
    linux_mapping.put()

    data_types.FuzzerJobs(platform='linux', fuzzer_jobs=[linux_mapping]).put()

    argument, job = fuzzer_selection.get_fuzz_task_payload('linux')
    self.assertEqual(('right_fuzzer', 'job_2'), (argument, job))


@test_utils.with_cloud_emulators('datastore')
class UpdatePlatformForJobTest(unittest.TestCase):
  """Tests for update_platform_for_job."""

  def test_jobs_updated(self):
    """Ensure that we properly update multiple jobs."""
    fuzzer_1_mapping = data_types.FuzzerJob()
    fuzzer_1_mapping.fuzzer = 'fuzzer_1'
    fuzzer_1_mapping.job = 'test_job'
    fuzzer_1_mapping.platform = 'wrong_platform'
    fuzzer_1_mapping.put()

    fuzzer_2_mapping = data_types.FuzzerJob()
    fuzzer_2_mapping.fuzzer = 'fuzzer_2'
    fuzzer_2_mapping.job = 'test_job'
    fuzzer_2_mapping.platform = 'wrong_platform'
    fuzzer_2_mapping.put()

    fuzzer_selection.update_platform_for_job('test_job', 'right_platform')

    platforms = [
        job.platform
        for job in ndb_utils.get_all_from_model(data_types.FuzzerJob)
    ]
    self.assertListEqual(platforms, ['right_platform', 'right_platform'])

  def test_unrelated_job_not_updated(self):
    """Ensure that we only update platform for the specified job's mappings."""
    fuzzer_1_mapping = data_types.FuzzerJob()
    fuzzer_1_mapping.fuzzer = 'fuzzer_1'
    fuzzer_1_mapping.job = 'test_job'
    fuzzer_1_mapping.platform = 'wrong_platform'
    fuzzer_1_mapping.put()

    fuzzer_2_mapping = data_types.FuzzerJob()
    fuzzer_2_mapping.fuzzer = 'fuzzer_2'
    fuzzer_2_mapping.job = 'unrelated_job'
    fuzzer_2_mapping.platform = 'unrelated_platform'
    fuzzer_2_mapping.put()

    fuzzer_selection.update_platform_for_job('test_job', 'right_platform')

    query = data_types.FuzzerJob.query()
    query = query.filter(data_types.FuzzerJob.fuzzer == 'fuzzer_2')
    fuzzer_2_mapping = query.get()

    self.assertEqual(fuzzer_2_mapping.platform, 'unrelated_platform')


class SelectFuzzTargetTest(unittest.TestCase):
  """Tests for select_fuzz_target."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.random_weighted_choice',
    ])

    self.mock.random_weighted_choice.side_effect = lambda x: x[0]

  def test_empty_weights(self):
    """Ensure that we select with even weights if none are specified."""
    fuzzer_selection.select_fuzz_target(['target_0', 'target_1'], {})

    expected_target_0 = fuzzer_selection.WeightedTarget('target_0', 1.0)
    expected_target_1 = fuzzer_selection.WeightedTarget('target_1', 1.0)
    self.mock.random_weighted_choice.assert_called_once_with(
        [expected_target_0, expected_target_1])

  def test_weights_for_some_targets(self):
    """Ensure that we use weights we applicable, and defaults where not."""
    fuzzer_selection.select_fuzz_target(
        ['weighted_0', 'weighted_1', 'unweighted_0'], {
            'weighted_0': 2.0,
            'weighted_1': 0.5
        })

    expected_weighted_0 = fuzzer_selection.WeightedTarget('weighted_0', 2.0)
    expected_weighted_1 = fuzzer_selection.WeightedTarget('weighted_1', 0.5)
    expected_unweighted_0 = fuzzer_selection.WeightedTarget('unweighted_0', 1.0)

    self.mock.random_weighted_choice.assert_called_once_with(
        [expected_weighted_0, expected_weighted_1, expected_unweighted_0])


@test_utils.with_cloud_emulators('datastore')
class GetFuzzTargetWeightsTest(unittest.TestCase):
  """Tests for get_fuzz_target_weights."""

  def setUp(self):
    test_helpers.patch_environ(self)

    data_types.FuzzTarget(
        engine='engine',
        project='proj',
        binary='child_0',
    ).put()
    data_types.FuzzTargetJob(
        fuzz_target_name='engine_proj_child_0',
        engine='engine',
        job='some_job_engine',
        weight=1.5,
    ).put()

    data_types.FuzzTarget(
        engine='engine',
        project='proj',
        binary='child_1',
    ).put()
    data_types.FuzzTargetJob(
        fuzz_target_name='engine_proj_child_1',
        engine='engine',
        job='some_job_engine',
        weight=0.75,
    ).put()

    data_types.FuzzTarget(
        engine='different_fuzzer',
        binary='child_2',
        project='test-project',
    ).put()
    data_types.FuzzTargetJob(
        fuzz_target_name='different_fuzzer_child_2',
        engine='different_fuzzer',
        job='some_job_different_fuzzer',
        weight=3.0,
    ).put()

  def test_empty_if_no_children(self):
    """Ensure that we function properly if a fuzzer has no children."""
    os.environ['TASK_NAME'] = 'fuzz'
    os.environ['JOB_NAME'] = 'some_job_blackbox'

    result = fuzzer_selection.get_fuzz_target_weights()

    self.assertEqual(result, {})

  def test_weights_specified_for_children(self):
    """Ensure that we return the correct weights for a fuzzer with children."""
    os.environ['TASK_NAME'] = 'fuzz'
    os.environ['JOB_NAME'] = 'some_job_engine'

    result = fuzzer_selection.get_fuzz_target_weights()

    self.assertEqual(result, {'child_0': 1.5, 'child_1': 0.75})



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""BigQuery client. We cannot use gcloud's BigQuery client
  because it requires oauth2client 4.0.0. But our appengine requires
  oauth2client 1.4.2. Therefore, we implement our own BigQuery client."""

import collections
import datetime
import re
import time

from googleapiclient import discovery

from clusterfuzz._internal.base import retry
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.google_cloud_utils import credentials
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

REQUEST_TIMEOUT = 60
QUERY_TIMEOUT = 10 * 60
QUERY_MAX_RESULTS = 10000
QUERY_RETRY_COUNT = 3
QUERY_RETRY_DELAY = 3

# The actual limit is 10,000, but the documentation suggests using batches of
# 500 rows at most (https://cloud.google.com/bigquery/quotas#streaminginserts).
INSERT_BATCH_SIZE = 500

VALID_FIELD_NAME_REGEX = re.compile(r'^[a-zA-Z0-9_]+$')


@retry.wrap(
    retries=QUERY_RETRY_COUNT,
    delay=QUERY_RETRY_DELAY,
    function='google_cloud_utils.big_query.get_api_client')
def get_api_client():
  """Return an api client for bigquery."""
  return discovery.build(
      'bigquery',
      'v2',
      cache_discovery=False,
      credentials=credentials.get_default()[0])


def get_bucket():
  """Return bucket for bigquery stats."""
  return local_config.ProjectConfig().get('bigquery.bucket')


def cast(value, field):
  """Cast value to appropriate type."""
  if value is None:
    return None

  if field['type'] in {'INTEGER', 'INT64'}:
    return int(value)
  if field['type'] in {'FLOAT', 'FLOAT64'}:
    return float(value)
  if field['type'] in {'BOOLEAN', 'BOOL'}:
    return value == 'true'
  if field['type'] in {'STRING'}:
    return value
  if field['type'] in {'TIMESTAMP'}:
    return datetime.datetime.utcfromtimestamp(float(value))
  if field['type'] in {'RECORD'}:
    return convert_row(value, field['fields'])
  raise Exception('The type %s is unsupported.' % field['type'])


def convert_row(raw_row, fields):
  """Convert a single raw row (from BigQuery) to a dict."""
  row = {}

  for index, raw_value in enumerate(raw_row['f']):
    field = fields[index]
    if field['mode'] == 'REPEATED':
      row[field['name']] = []
      for item in raw_value['v']:
        row[field['name']].append(cast(item['v'], field))
    else:
      row[field['name']] = cast(raw_value['v'], field)

  return row


def convert(result):
  """Convert a query result into an array of dicts, each of which represents
    a row."""
  fields = result['schema']['fields']
  rows = []

  for raw_row in result.get('rows', []):
    rows.append(convert_row(raw_row, fields))

  return rows


@environment.local_noop
def write_range(table_id, testcase, range_name, start, end):
  """Write a range to BigQuery. This is applicable for regression and fixed
    ranges."""
  client = Client(dataset_id='main', table_id=table_id)
  result = client.insert([
      Insert(
          row={
              'testcase_id': str(testcase.key.id()),
              'crash_type': testcase.crash_type,
              'crash_state': testcase.crash_state,
              'security_flag': testcase.security_flag,
              'parent_fuzzer_name': testcase.fuzzer_name,
              'fuzzer_name': testcase.overridden_fuzzer_name,
              'job_type': testcase.job_type,
              'created_at': int(time.time()),
              ('%s_range_start' % range_name): int(start),
              ('%s_range_end' % range_name): int(end),
          },
          insert_id='%s:%s:%s' % (testcase.key.id(), start, end))
  ])

  for error in result.get('insertErrors', []):
    logs.log_error(
        ("Ignoring error writing the testcase's %s range (%s) to "
         'BigQuery.' % (range_name, testcase.key.id())),
        exception=Exception(error))


def _get_max_results(max_results, limit, count_so_far):
  """Get an appropriate max_results."""
  # limit is None means we get every record (no limit).
  if limit is None:
    return max_results

  return min(max_results, limit - count_so_far)


Insert = collections.namedtuple('Insert', ['row', 'insert_id'])
QueryResult = collections.namedtuple('QueryResult', ['rows', 'total_count'])


class Client(object):
  """BigQuery client."""

  def __init__(self, dataset_id=None, table_id=None):
    self.project_id = utils.get_application_id()
    self.dataset_id = dataset_id
    self.table_id = table_id

    self.client = get_api_client()

  @retry.wrap(
      retries=QUERY_RETRY_COUNT,
      delay=QUERY_RETRY_DELAY,
      function='google_cloud_utils.big_query.Client.raw_query')
  def raw_query(self, query, max_results):
    # pylint: disable=line-too-long
    """Perform a query and return result.

    Args:
      query: the query string.
      timeout: the timeout in seconds.
      max_results: the number of rows per response. The response cannot exceed
        10MB.
      use_legacy_sql: whether or not the query is of the legacy sql.

    Returns:
      A json explained here:
      https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/query
    """
    body = {
        'query': query,
        'timeoutMs': REQUEST_TIMEOUT * 1000,
        'useLegacySql': False,
        'maxResults': max_results
    }

    return self.client.jobs().query(
        projectId=self.project_id, body=body).execute()

  @retry.wrap(
      retries=QUERY_RETRY_COUNT,
      delay=QUERY_RETRY_DELAY,
      function='google_cloud_utils.big_query.Client.get_query_results')
  def get_query_results(self, job_id, page_token, start_index, max_results):
    # pylint: disable=line-too-long
    """Perform a query and return result.

    Args:
      query: the query string.
      job_id: the job id from query's response.
      page_token: the page token from the previous query's response.
      max_results: the number of rows per response. The response cannot exceed
        10MB.

    Returns:
      A json explained here:
      https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/getQueryResults
    """
    return self.client.jobs().getQueryResults(
        projectId=self.project_id,
        jobId=job_id,
        timeoutMs=REQUEST_TIMEOUT * 1000,
        maxResults=max_results,
        startIndex=start_index,
        pageToken=page_token).execute()

  def wait_for_completion(self, job_id, offset, max_results, start_time,
                          timeout):
    """Wait for job completion and return the first page."""
    while True:
      result = self.get_query_results(
          job_id=job_id,
          page_token=None,
          start_index=offset,
          max_results=max_results)

      if result['jobComplete']:
        return result

      if (time.time() - start_time) > timeout:
        raise Exception(
            "Timeout: the query doesn't finish within %d seconds." % timeout)
      time.sleep(1)

  def query(self,
            query,
            timeout=QUERY_TIMEOUT,
            max_results=QUERY_MAX_RESULTS,
            offset=0,
            limit=None):
    """Performs a query and returns an array of dicts."""
    rows = []
    start_time = time.time()

    result = self.raw_query(query, max_results=0)

    result = self.wait_for_completion(
        job_id=result['jobReference']['jobId'],
        offset=offset,
        max_results=_get_max_results(max_results, limit, 0),
        start_time=start_time,
        timeout=timeout)

    # totalRows is only present after the job completed successfully.
    total_count = int(result['totalRows'])

    while limit is None or len(rows) < limit:
      rows += convert(result)

      if result['jobComplete'] and 'pageToken' not in result:
        total_count = int(result['totalRows'])
        break

      result = self.get_query_results(
          job_id=result['jobReference']['jobId'],
          page_token=result.get('pageToken'),
          start_index=0,
          max_results=_get_max_results(max_results, limit, len(rows)))

    return QueryResult(rows=rows, total_count=total_count)

  @retry.wrap(
      retries=QUERY_RETRY_COUNT,
      delay=QUERY_RETRY_DELAY,
      function='google_cloud_utils.big_query.Client.get_job')
  def get_job(self, job_id):
    # pylint: disable=line-too-long
    """Get the job.

    Args:
      job_id: the job id.

    Returns:
      A json explained here:
      https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs#configuration.query
    """
    return self.client.jobs().get(
        projectId=self.project_id, jobId=job_id).execute()

  @environment.local_noop
  @retry.wrap(
      retries=QUERY_RETRY_COUNT,
      delay=QUERY_RETRY_DELAY,
      function='google_cloud_utils.big_query.Client.insert_from_query')
  def insert_from_query(self, dataset_id, table_id, job_id, query):
    # pylint: disable=line-too-long
    """Insert rows to the table from a query.

    Args:
      dataset_id: the destination dataset id.
      table_id: the desitnation table id.
      job_id: the uniquely identified job id (used for preventing redundant
        job).
      query: the query that generates rows.

    Returns:
      A json explained here:
      https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs#configuration.query
    """
    return self.client.jobs().insert(
        projectId=self.project_id,
        body={
            'configuration': {
                'query': {
                    'query': query,
                    'allowLargeResults': True,
                    'destinationTable': {
                        'projectId': self.project_id,
                        'datasetId': dataset_id,
                        'tableId': table_id
                    },
                    'useLegacySql': False,
                    'writeDisposition': 'WRITE_APPEND'
                }
            },
            'jobReference': {
                'jobId': job_id,
                'projectId': self.project_id
            }
        }).execute()

  @environment.local_noop
  @retry.wrap(
      retries=QUERY_RETRY_COUNT,
      delay=QUERY_RETRY_DELAY,
      function='google_cloud_utils.big_query.Client.insert')
  def _insert_batch(self, inserts):
    # pylint: disable=line-too-long
    """Insert a single batch of rows to BigQuery.

    Args:
      inserts: a list of Inserts, each of which represents a row.

    Returns:
      A json explained here:
      https://cloud.google.com/bigquery/docs/reference/rest/v2/tabledata/insertAll
    """
    inserted_rows = []
    for insert in inserts:
      inserted_rows.append({'json': insert.row, 'insertId': insert.insert_id})

    body = {'kind': 'bigquery#tableDataInsertAllRequest', 'rows': inserted_rows}
    return self.client.tabledata().insertAll(
        projectId=self.project_id,
        datasetId=self.dataset_id,
        tableId=self.table_id,
        body=body).execute()

  @environment.local_noop
  def insert(self, inserts):
    # pylint: disable=line-too-long
    """Insert the given rows and batch them if needed.

    Args:
        inserts: a list of Inserts, each of which represents a row.

    Returns:
      A json explained here:
      https://cloud.google.com/bigquery/docs/reference/rest/v2/tabledata/insertAll
    """
    result = {}
    for i in range(0, len(inserts), INSERT_BATCH_SIZE):
      response = self._insert_batch(
          inserts[i:min(len(inserts), i + INSERT_BATCH_SIZE)])

      # There is also quota for the total number of requests / rows per second.
      time.sleep(1)

      if not result:
        # Use result from the first batch, appending errors from the rest.
        result = response
        continue

      # If there are new errors from the current batch, append to the result.
      new_errors = response.get('insertErrors')
      if not new_errors:
        continue

      # Apparently result may not have errors, use |setdefault| to be careful.
      result.setdefault('insertErrors', []).extend(new_errors)

    return result
